#!/usr/bin/env python3

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2017 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.12.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

from anacore.sequenceIO import IdxFastaIO
import os
import pysam
import sys
import tempfile
import unittest
import uuid

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
PACKAGE_DIR = os.path.dirname(TEST_DIR)
sys.path.append(PACKAGE_DIR)

from anacore.vcf import VCFRecord, VCFIO, getHeaderAttr, HeaderInfoAttr


########################################################################
#
# FUNCTIONS
#
########################################################################
class TestVCFIO(unittest.TestCase):
    def setUp(self):
        tmp_folder = tempfile.gettempdir()
        unique_id = str(uuid.uuid1())

        # Temporary files
        self.tmp_in_variants = os.path.join(tmp_folder, unique_id + "_in.vcf")
        self.tmp_in_variants_with_spl_info = os.path.join(tmp_folder, unique_id + "_in_detailed.vcf")
        self.tmp_out_variants = os.path.join(tmp_folder, unique_id + "_out.vcf")

        # Create VCF
        spec_example = """##fileformat=VCFv4.3
##fileDate=20090805"
##source=myImputationProgramV3.1
##reference=file:///seq/references/1000GenomesPilot-NCBI36.fasta
##contig=<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,species="Homo sapiens",taxonomy=x>
##phasing=partial
##INFO=<ID=NS,Number=1,Type=Integer,Description="Number of Samples With Data">
##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
##INFO=<ID=AF,Number=A,Type=Float,Description="Allele Frequency">
##INFO=<ID=AA,Number=1,Type=String,Description="Ancestral Allele">
##INFO=<ID=DB,Number=0,Type=Flag,Description="dbSNP membership, build 129">
##INFO=<ID=H2,Number=0,Type=Flag,Description="HapMap2 membership">
##FILTER=<ID=q10,Description="Quality below 10">
##FILTER=<ID=s50,Description="Less than 50% of samples have data">
##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
##FORMAT=<ID=GQ,Number=1,Type=Integer,Description="Genotype Quality">
##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read Depth">
##FORMAT=<ID=HQ,Number=2,Type=Integer,Description="Haplotype Quality">
#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT	NA00001	NA00002	NA00003
20	14370	rs6054257	G	A	29	PASS	NS=3;DP=14;AF=0.5;DB;H2	GT:GQ:DP:HQ	0|0:48:1:51,51	1|0:48:8:51,51	1/1:43:5:.,.
20	17330	.	T	A	3	q10	NS=3;DP=11;AF=0.017	GT:GQ:DP:HQ	0|0:49:3:58,50	0|1:3:5:65,3	0/0:41:3
20	1110696	rs6040355	A	G,T	67	PASS	NS=2;DP=10;AF=0.333,0.667;AA=T;DB	GT:GQ:DP:HQ	1|2:21:6:23,27	2|1:2:0:18,2	2/2:35:4
20	1230237	.	T	-	47	PASS	NS=3;DP=13;AA=T	GT:GQ:DP:HQ	0|0:54:7:56,60	0|0:48:4:51,51	0/0:61:2
20	1234567	microsat1	GTC	G,GTCT	50	PASS	NS=3;DP=9;AA=G	GT:GQ:DP	0/1:35:4	0/2:17:2	1/1:40:3"""
        with open(self.tmp_in_variants, "w") as FH_variants:
            FH_variants.write(spec_example)

        # Create VCF with spl info
        spec_example_with_spl_info = spec_example.replace(
            "#CHROM",
            """##SAMPLE=<ID=NA00001,Assay=WholeGenome,Ethnicity=AFR,Disease=None,Description="Patient germline genome from unaffected">
##SAMPLE=<ID=NA00003,Assay=Exome,Ethnicity=CEU,Description="European patient exome from breast cancer",Disease=Cancer,Tissue=Breast>
#CHROM""")
        with open(self.tmp_in_variants_with_spl_info, "w") as FH_variants:
            FH_variants.write(spec_example_with_spl_info)

    def testIter(self):
        with VCFIO(self.tmp_in_variants) as FH_vcf:
            # Header
            self.assertEqual(FH_vcf.samples, ["NA00001", "NA00002", "NA00003"])
            self.assertEqual(sorted(list(FH_vcf.format.keys())), sorted(["GT", "GQ", "DP", "HQ"]))
            self.assertEqual(sorted(list(FH_vcf.info.keys())), sorted(["NS", "DP", "AF", "AA", "DB", "H2"]))
            # Records
            expected_records = ["20_14370_G_A", "20_17330_T_A", "20_1110696_A_G,T", "20_1230237_T_-", "20_1234567_GTC_G,GTCT"]
            readed_records = list()
            for variant in FH_vcf:
                readed_records.append(
                    "_".join([variant.chrom, str(variant.pos), variant.ref, ",".join(variant.alt)])
                )
            self.assertEqual(expected_records, readed_records)

    def testGetSub(self):
        try:
            pysam.tabix_compress(self.tmp_in_variants, self.tmp_in_variants + ".gz")
            pysam.tabix_index(self.tmp_in_variants + ".gz", preset="vcf")
            with VCFIO(self.tmp_in_variants + ".gz", "i") as reader:
                # Missing chromosome
                observed = [elt.getName() for elt in reader.getSub("1", 1437, 11437)]
                self.assertEqual(observed, [])
                # No variants
                observed = [elt.getName() for elt in reader.getSub("20", 1234580, 1234590)]
                self.assertEqual(observed, [])
                # At substit
                observed = [elt.getName() for elt in reader.getSub("20", 17330, 17330)]
                self.assertEqual(observed, ["20:17330=T/A"])
                # Contains several variants
                observed = [elt.getName() for elt in reader.getSub("20", 14330, 17350)]
                self.assertEqual(observed, ["20:14370=G/A", "20:17330=T/A"])
                # Overlap deletion start
                observed = [elt.getName() for elt in reader.getSub("20", 1234550, 1234567)]
                self.assertEqual(observed, ["20:1234567=GTC/G/GTCT"])
                # In deletion
                observed = [elt.getName() for elt in reader.getSub("20", 1234569, 1234569)]
                self.assertEqual(observed, ["20:1234567=GTC/G/GTCT"])
        finally:
            for curr_file in [self.tmp_in_variants + ".gz", self.tmp_in_variants + ".gz.tbi"]:
                if os.path.exists(curr_file):
                    os.remove(curr_file)

    def testParseHeader(self):
        with VCFIO(self.tmp_in_variants_with_spl_info) as FH_vcf:
            # FORMAT
            self.assertEqual(
                {key: elt.datastore for key, elt in FH_vcf.format.items()},
                {
                    'GT': {'id': 'GT', 'description': 'Genotype', 'type': 'String', 'number': '1'},
                    'GQ': {'id': 'GQ', 'description': 'Genotype Quality', 'type': 'Integer', 'number': '1'},
                    'DP': {'id': 'DP', 'description': 'Read Depth', 'type': 'Integer', 'number': '1'},
                    'HQ': {'id': 'HQ', 'description': 'Haplotype Quality', 'type': 'Integer', 'number': '2'}
                }
            )
            # INFO
            self.assertEqual(
                {key: elt.datastore for key, elt in FH_vcf.info.items()},
                {
                    'NS': {'id': 'NS', 'description': 'Number of Samples With Data', 'type': 'Integer', 'number': '1'},
                    'DP': {'id': 'DP', 'description': 'Total Depth', 'type': 'Integer', 'number': '1'},
                    'AF': {'id': 'AF', 'description': 'Allele Frequency', 'type': 'Float', 'number': 'A'},
                    'AA': {'id': 'AA', 'description': 'Ancestral Allele', 'type': 'String', 'number': '1'},
                    'DB': {'id': 'DB', 'description': 'dbSNP membership, build 129', 'type': 'Flag', 'number': '0'},
                    'H2': {'id': 'H2', 'description': 'HapMap2 membership', 'type': 'Flag', 'number': '0'}}
            )
            # Samples list
            self.assertEqual(FH_vcf.samples, ["NA00001", "NA00002", "NA00003"])
            # SAMPLE
            self.assertDictEqual(
                {key: elt.datastore for key, elt in FH_vcf.sample_info.items()},
                {
                    "NA00001": {
                        "id": "NA00001",
                        "assay": "WholeGenome",
                        "ethnicity": "AFR",
                        "disease": "None",
                        "description": "Patient germline genome from unaffected"
                    },
                    "NA00003": {
                        "id": "NA00003",
                        "assay": "Exome",
                        "ethnicity": "CEU",
                        "disease": "Cancer",
                        "description": "European patient exome from breast cancer",
                        "tissue": "Breast"
                    }
                }
            )
            # EXTRA
            self.assertEqual(
                FH_vcf.extra_header,
                [
                    '##fileDate=20090805"',
                    '##source=myImputationProgramV3.1',
                    '##reference=file:///seq/references/1000GenomesPilot-NCBI36.fasta',
                    '##contig=<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,species="Homo sapiens",taxonomy=x>',
                    '##phasing=partial'
                ]
            )

    def testReadTypeFlag(self):
        expected = [
            {"DB": True, "H2": True},
            {"DB": False, "H2": False},
            {"DB": True, "H2": False},
            {"DB": False, "H2": False},
            {"DB": False, "H2": False}
        ]
        observed = []
        with VCFIO(self.tmp_in_variants) as FH_vcf:
            for record in FH_vcf:
                observed.append({
                    "DB": "DB" in record.info,
                    "H2": "H2" in record.info
                })
        self.assertEqual(expected, observed)

    def testWrite(self):
        expected_content = """##fileformat=VCFv4.3
##fileDate=20090805"
##source=myImputationProgramV3.1
##reference=file:///seq/references/1000GenomesPilot-NCBI36.fasta
##contig=<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,species="Homo sapiens",taxonomy=x>
##phasing=partial
##INFO=<ID=AA,Number=1,Type=String,Description="Ancestral Allele">
##INFO=<ID=AF,Number=A,Type=Float,Description="Allele Frequency">
##INFO=<ID=DB,Number=0,Type=Flag,Description="dbSNP membership, build 129">
##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
##INFO=<ID=H2,Number=0,Type=Flag,Description="HapMap2 membership">
##INFO=<ID=NS,Number=1,Type=Integer,Description="Number of Samples With Data">
##FILTER=<ID=q10,Description="Quality below 10">
##FILTER=<ID=s50,Description="Less than 50% of samples have data">
##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read Depth">
##FORMAT=<ID=GQ,Number=1,Type=Integer,Description="Genotype Quality">
##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
##FORMAT=<ID=HQ,Number=2,Type=Integer,Description="Haplotype Quality">
#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT	NA00001	NA00002	NA00003
20	14370	rs6054257	G	A	29.0	PASS	AF=0.5;DB;DP=14;H2;NS=3	GT:GQ:DP:HQ	0|0:48:1:51,51	1|0:48:8:51,51	1/1:43:5:.,.
20	17330	.	T	A	3.0	q10	AF=0.017;DP=11;NS=3	GT:GQ:DP:HQ	0|0:49:3:58,50	0|1:3:5:65,3	0/0:41:3:.
20	1110696	rs6040355	A	G,T	67.0	PASS	AA=T;AF=0.333,0.667;DB;DP=10;NS=2	GT:GQ:DP:HQ	1|2:21:6:23,27	2|1:2:0:18,2	2/2:35:4:.
20	1230237	.	T	-	47.0	PASS	AA=T;DP=13;NS=3	GT:GQ:DP:HQ	0|0:54:7:56,60	0|0:48:4:51,51	0/0:61:2:.
20	1234567	microsat1	GTC	G,GTCT	50.0	PASS	AA=G;DP=9;NS=3	GT:GQ:DP	0/1:35:4	0/2:17:2	1/1:40:3"""
        # Read and write VCF
        with VCFIO(self.tmp_in_variants) as reader:
            with VCFIO(self.tmp_out_variants, "w") as writer:
                writer.copyHeader(reader)
                writer.writeHeader()
                for record in reader:
                    writer.write(record)
        # Compare input content with output
        observed_content = None
        with open(self.tmp_out_variants) as reader:
            observed_content = "".join(reader.readlines()).strip()
        self.assertEqual(expected_content, observed_content)

    def testWriteWithSplInfo(self):
        expected_content = """##fileformat=VCFv4.3
##fileDate=20090805"
##source=myImputationProgramV3.1
##reference=file:///seq/references/1000GenomesPilot-NCBI36.fasta
##contig=<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,species="Homo sapiens",taxonomy=x>
##phasing=partial
##INFO=<ID=AA,Number=1,Type=String,Description="Ancestral Allele">
##INFO=<ID=AF,Number=A,Type=Float,Description="Allele Frequency">
##INFO=<ID=DB,Number=0,Type=Flag,Description="dbSNP membership, build 129">
##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
##INFO=<ID=H2,Number=0,Type=Flag,Description="HapMap2 membership">
##INFO=<ID=NS,Number=1,Type=Integer,Description="Number of Samples With Data">
##FILTER=<ID=q10,Description="Quality below 10">
##FILTER=<ID=s50,Description="Less than 50% of samples have data">
##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read Depth">
##FORMAT=<ID=GQ,Number=1,Type=Integer,Description="Genotype Quality">
##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
##FORMAT=<ID=HQ,Number=2,Type=Integer,Description="Haplotype Quality">
##SAMPLE=<ID=NA00001,Assay="WholeGenome",Description="Patient germline genome from unaffected",Disease="None",Ethnicity="AFR">
##SAMPLE=<ID=NA00003,Assay="Exome",Description="European patient exome from breast cancer",Disease="Cancer",Ethnicity="CEU",Tissue="Breast">
#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT	NA00001	NA00002	NA00003
20	14370	rs6054257	G	A	29.0	PASS	AF=0.5;DB;DP=14;H2;NS=3	GT:GQ:DP:HQ	0|0:48:1:51,51	1|0:48:8:51,51	1/1:43:5:.,.
20	17330	.	T	A	3.0	q10	AF=0.017;DP=11;NS=3	GT:GQ:DP:HQ	0|0:49:3:58,50	0|1:3:5:65,3	0/0:41:3:.
20	1110696	rs6040355	A	G,T	67.0	PASS	AA=T;AF=0.333,0.667;DB;DP=10;NS=2	GT:GQ:DP:HQ	1|2:21:6:23,27	2|1:2:0:18,2	2/2:35:4:.
20	1230237	.	T	-	47.0	PASS	AA=T;DP=13;NS=3	GT:GQ:DP:HQ	0|0:54:7:56,60	0|0:48:4:51,51	0/0:61:2:.
20	1234567	microsat1	GTC	G,GTCT	50.0	PASS	AA=G;DP=9;NS=3	GT:GQ:DP	0/1:35:4	0/2:17:2	1/1:40:3"""
        # Read and write VCF
        with VCFIO(self.tmp_in_variants_with_spl_info) as reader:
            with VCFIO(self.tmp_out_variants, "w") as writer:
                writer.copyHeader(reader)
                writer.writeHeader()
                for record in reader:
                    writer.write(record)
        # Compare input content with output
        observed_content = None
        with open(self.tmp_out_variants) as reader:
            observed_content = "".join(reader.readlines()).strip()
        self.assertEqual(expected_content, observed_content)

    def tearDown(self):
        # Clean temporary files
        for curr_file in [self.tmp_in_variants, self.tmp_in_variants_with_spl_info, self.tmp_out_variants]:
            if os.path.exists(curr_file):
                os.remove(curr_file)


class TestVCFRecord(unittest.TestCase):
    def setUp(self):
        tmp_folder = tempfile.gettempdir()
        unique_id = str(uuid.uuid1())

        # Temporary files
        self.tmp_fa = os.path.join(tmp_folder, unique_id + ".fa")
        self.tmp_fai = os.path.join(tmp_folder, unique_id + ".fa.fai")
        self.tmp_variants = os.path.join(tmp_folder, unique_id + ".vcf")

        # AD, AF and DP evaluation
        self.freq_data = dict()
        self.freq_data["wout_spl_rAD_aAF"] = """##fileformat=VCFv4.0
##INFO=<ID=AD,Number=R,Type=Integer,Description="Allele Depth">
##INFO=<ID=AF,Number=A,Type=Float,Description="Allele Frequency">
##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
##INFO=<ID=expModel,Number=1,Type=String,Description="Expected model for variant">
#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT
11	1	wout_spl_rAD_aAF_1	A	T	29	PASS	expModel=model_0;AD=90,10;DP=100	.
11	2	wout_spl_rAD_aAF_2	A	T	29	PASS	expModel=model_0;AF=0.1;AD=90,10	.
11	3	wout_spl_rAD_aAF_3	A	T	29	PASS	expModel=model_0;AF=0.1;DP=100	.
11	1	wout_spl_rAD_aAF_4	A	T,G	29	PASS	expModel=model_1;AD=85,10,5;DP=100	.
11	2	wout_spl_rAD_aAF_5	A	T,G	29	PASS	expModel=model_1;AF=0.1,0.05;AD=85,10,5	.
11	3	wout_spl_rAD_aAF_6	A	T,G	29	PASS	expModel=model_1;AF=0.1,0.05;DP=100	."""
        self.freq_data["wout_spl_aAD_rAF"] = """##fileformat=VCFv4.0
##INFO=<ID=AD,Number=A,Type=Integer,Description="Allele Depth">
##INFO=<ID=AF,Number=R,Type=Float,Description="Allele Frequency">
##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
##INFO=<ID=expModel,Number=1,Type=String,Description="Expected model for variant">
#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT
11	1	wout_spl_aAD_rAF_1	A	T	29	PASS	expModel=model_0;AD=10;DP=100	.
11	2	wout_spl_aAD_rAF_2	A	T	29	PASS	expModel=model_0;AF=0.9,0.1;AD=10	.
11	3	wout_spl_aAD_rAF_3	A	T	29	PASS	expModel=model_0;AF=0.9,0.1;DP=100	.
11	1	wout_spl_aAD_rAF_4	A	T,G	29	PASS	expModel=model_1;AD=10,5;DP=100	.
11	2	wout_spl_aAD_rAF_5	A	T,G	29	PASS	expModel=model_1;AF=0.85,0.1,0.05;AD=10,5	.
11	3	wout_spl_aAD_rAF_6	A	T,G	29	PASS	expModel=model_1;AF=0.85,0.1,0.05;DP=100	."""
        self.freq_data["one_spl_rAD_aAF"] = """##fileformat=VCFv4.0
##INFO=<ID=AD,Number=R,Type=Integer,Description="Allele Depth">
##INFO=<ID=AF,Number=A,Type=Float,Description="Allele Frequency">
##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
##INFO=<ID=expModel,Number=1,Type=String,Description="Expected model for variant">
##INFO=<ID=skipAD,Number=1,Type=String,Description="Variant must be skipped by AD calculation">
##INFO=<ID=skipAF,Number=1,Type=String,Description="Variant must be skipped by AF calculation">
##INFO=<ID=skipDP,Number=1,Type=String,Description="Variant must be skipped by DP calculation">
##FORMAT=<ID=AD,Number=R,Type=Integer,Description="Allele Depth">
##FORMAT=<ID=AF,Number=A,Type=Float,Description="Allele Frequency">
##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT	splA
11	1	one_spl_rAD_aAF_1	A	T	29	PASS	expModel=model_2;skipAF=True;AD=90,10;DP=100	.	.
11	2	one_spl_rAD_aAF_2	A	T	29	PASS	expModel=model_2;skipDP=True;AF=0.1;AD=90,10	.	.
11	3	one_spl_rAD_aAF_3	A	T	29	PASS	expModel=model_2;skipAD=True;AF=0.1;DP=100	.	.
11	1	one_spl_rAD_aAF_4	A	T	29	PASS	expModel=model_2	AD:DP	90,10:100
11	2	one_spl_rAD_aAF_5	A	T	29	PASS	expModel=model_2	AF:AD	0.1:90,10
11	3	one_spl_rAD_aAF_6	A	T	29	PASS	expModel=model_2	AF:DP	0.1:100
11	1	one_spl_rAD_aAF_7	A	T,G	29	PASS	expModel=model_3;skipAF=True;AD=85,10,5;DP=100	.	.
11	2	one_spl_rAD_aAF_8	A	T,G	29	PASS	expModel=model_3;skipDP=True;AF=0.1,0.05;AD=85,10,5	.	.
11	3	one_spl_rAD_aAF_9	A	T,G	29	PASS	expModel=model_3;skipAD=True;AF=0.1,0.05;DP=100	.	.
11	1	one_spl_rAD_aAF_10	A	T,G	29	PASS	expModel=model_3	AD:DP	85,10,5:100
11	2	one_spl_rAD_aAF_11	A	T,G	29	PASS	expModel=model_3	AF:AD	0.1,0.05:85,10,5
11	3	one_spl_rAD_aAF_12	A	T,G	29	PASS	expModel=model_3	AF:DP	0.1,0.05:100"""
        self.freq_data["one_spl_aAD_rAF"] = """##fileformat=VCFv4.0
##INFO=<ID=AD,Number=A,Type=Integer,Description="Allele Depth">
##INFO=<ID=AF,Number=R,Type=Float,Description="Allele Frequency">
##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
##INFO=<ID=expModel,Number=1,Type=String,Description="Expected model for variant">
##INFO=<ID=skipAD,Number=1,Type=String,Description="Variant must be skipped by AD calculation">
##INFO=<ID=skipAF,Number=1,Type=String,Description="Variant must be skipped by AF calculation">
##INFO=<ID=skipDP,Number=1,Type=String,Description="Variant must be skipped by DP calculation">
##FORMAT=<ID=AD,Number=A,Type=Integer,Description="Allele Depth">
##FORMAT=<ID=AF,Number=R,Type=Float,Description="Allele Frequency">
##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT	splA
11	1	one_spl_aAD_rAF_1	A	T	29	PASS	expModel=model_2;skipAF=True;AD=10;DP=100	.	.
11	2	one_spl_aAD_rAF_2	A	T	29	PASS	expModel=model_2;skipDP=True;AF=0.9,0.1;AD=10	.	.
11	3	one_spl_aAD_rAF_3	A	T	29	PASS	expModel=model_2;skipAD=True;AF=0.9,0.1;DP=100	.	.
11	1	one_spl_aAD_rAF_4	A	T	29	PASS	expModel=model_2	AD:DP	10:100
11	2	one_spl_aAD_rAF_5	A	T	29	PASS	expModel=model_2	AF:AD	0.9,0.1:10
11	3	one_spl_aAD_rAF_6	A	T	29	PASS	expModel=model_2	AF:DP	0.9,0.1:100
11	1	one_spl_rAD_aAF_7	A	T,G	29	PASS	expModel=model_3;skipAF=True;AD=10,5;DP=100	.	.
11	2	one_spl_rAD_aAF_8	A	T,G	29	PASS	expModel=model_3;skipDP=True;AF=0.85,0.1,0.05;AD=10,5	.	.
11	3	one_spl_rAD_aAF_9	A	T,G	29	PASS	expModel=model_3;skipAD=True;AF=0.85,0.1,0.05;DP=100	.	.
11	1	one_spl_rAD_aAF_10	A	T,G	29	PASS	expModel=model_3	AD:DP	10,5:100
11	2	one_spl_rAD_aAF_11	A	T,G	29	PASS	expModel=model_3	AF:AD	0.85,0.1,0.05:10,5
11	3	one_spl_rAD_aAF_12	A	T,G	29	PASS	expModel=model_3	AF:DP	0.85,0.1,0.05:100"""
        self.freq_data["two_spl_rAD_aAF"] = """##fileformat=VCFv4.0
##INFO=<ID=AD,Number=A,Type=Integer,Description="Allele Depth">
##INFO=<ID=AF,Number=R,Type=Float,Description="Allele Frequency">
##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
##INFO=<ID=expModel,Number=1,Type=String,Description="Expected model for variant">
##INFO=<ID=skipAD,Number=1,Type=String,Description="Variant must be skipped by AD calculation">
##INFO=<ID=skipAF,Number=1,Type=String,Description="Variant must be skipped by AF calculation">
##INFO=<ID=skipDP,Number=1,Type=String,Description="Variant must be skipped by DP calculation">
##FORMAT=<ID=AD,Number=A,Type=Integer,Description="Allele Depth">
##FORMAT=<ID=AF,Number=R,Type=Float,Description="Allele Frequency">
##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT	splA	splB
11	1	two_spl_rAD_aAF_1	A	T	29	PASS	expModel=model_4	AD:DP	48,2:50	42,8:50
11	2	two_spl_rAD_aAF_2	A	T	29	PASS	expModel=model_4	AF:AD	0.04:48,2	0.16:42,8
11	3	two_spl_rAD_aAF_3	A	T	29	PASS	expModel=model_4	AF:DP	0.04:50	0.16:50
11	1	two_spl_rAD_aAF_4	A	T,G	29	PASS	expModel=model_5	AD:DP	43,2,5:50	41,8,1:50
11	2	two_spl_rAD_aAF_5	A	T,G	29	PASS	expModel=model_5	AF:AD	0.04,0.1:43,2,5	0.16,0.02:41,8,1
11	3	two_spl_rAD_aAF_6	A	T,G	29	PASS	expModel=model_5	AF:DP	0.04,0.1:50	0.16,0.02:50"""
        self.freq_data["two_spl_aAD_rAF"] = """##fileformat=VCFv4.0
##INFO=<ID=AD,Number=A,Type=Integer,Description="Allele Depth">
##INFO=<ID=AF,Number=R,Type=Float,Description="Allele Frequency">
##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
##INFO=<ID=expModel,Number=1,Type=String,Description="Expected model for variant">
##INFO=<ID=skipAD,Number=1,Type=String,Description="Variant must be skipped by AD calculation">
##INFO=<ID=skipAF,Number=1,Type=String,Description="Variant must be skipped by AF calculation">
##INFO=<ID=skipDP,Number=1,Type=String,Description="Variant must be skipped by DP calculation">
##FORMAT=<ID=AD,Number=A,Type=Integer,Description="Allele Depth">
##FORMAT=<ID=AF,Number=R,Type=Float,Description="Allele Frequency">
##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT	splA	splB
11	1	two_spl_rAD_aAF_1	A	T	29	PASS	expModel=model_4	AD:DP	2:50	8:50
11	2	two_spl_rAD_aAF_2	A	T	29	PASS	expModel=model_4	AF:AD	0.96,0.04:2	0.84,0.16:8
11	3	two_spl_rAD_aAF_3	A	T	29	PASS	expModel=model_4	AF:DP	0.96,0.04:50	0.84,0.16:50
11	1	two_spl_rAD_aAF_4	A	T,G	29	PASS	expModel=model_5	AD:DP	2,5:50	8,1:50
11	2	two_spl_rAD_aAF_5	A	T,G	29	PASS	expModel=model_5	AF:AD	0.96,0.04,0.1:2,5	0.84,0.16,0.02:8,1
11	3	two_spl_rAD_aAF_6	A	T,G	29	PASS	expModel=model_5	AF:DP	0.96,0.04,0.1:50	0.84,0.16,0.02:50"""
        self.freq_expected = {
            "model_0": {
                "AF": {"pop": [0.1]},
                "AD": {"pop": [10]},
                "DP": {"pop": 100}
            },
            "model_1": {
                "AF": {"pop": [0.1, 0.05]},
                "AD": {"pop": [10, 5]},
                "DP": {"pop": 100}
            },
            "model_2": {
                "AF": {"splA": [0.1], "pop": [0.1]},
                "AD": {"splA": [10], "pop": [10]},
                "DP": {"splA": 100, "pop": 100}
            },
            "model_3": {
                "AF": {"splA": [0.1, 0.05], "pop": [0.1, 0.05]},
                "AD": {"splA": [10, 5], "pop": [10, 5]},
                "DP": {"splA": 100, "pop": 100}
            },
            "model_4": {
                "AF": {"splA": [0.04], "splB": [0.16], "pop": [0.1]},
                "AD": {"splA": [2], "splB": [8], "pop": [10]},
                "DP": {"splA": 50, "splB": 50, "pop": 100}
            },
            "model_5": {
                "AF": {"splA": [0.04, 0.1], "splB": [0.16, 0.02], "pop": [0.1, 0.06]},
                "AD": {"splA": [2, 5], "splB": [8, 1], "pop": [10, 6]},
                "DP": {"splA": 50, "splB": 50, "pop": 100}
            }
        }

    def tearDown(self):
        # Clean temporary files
        for curr_file in [self.tmp_fa, self.tmp_fai, self.tmp_variants]:
            if os.path.exists(curr_file):
                os.remove(curr_file)

    def testGetAltAF(self):
        for curr_dataset in sorted(self.freq_data):
            if not curr_dataset.startswith("wout_spl"):
                # Write test data
                content = self.freq_data[curr_dataset]
                with open(self.tmp_variants, "w") as FH_variants:
                    FH_variants.write(content)
                # Parse
                expected_records = list()
                observed_records = list()
                with VCFIO(self.tmp_variants) as FH_vcf:
                    for variant in FH_vcf:
                        if "skipAF" not in variant.info:
                            try:
                                for idx_spl, curr_spl in enumerate(FH_vcf.samples):
                                    expected_AF = self.freq_expected[variant.info["expModel"]]["AF"][curr_spl]
                                    expected_records.append("{} {}".format(variant.id, expected_AF))
                                    observed_records.append("{} {}".format(variant.id, variant.getAltAF(curr_spl)))
                            except Exception:
                                raise Exception('Error in TestVCFRecord.testGetAltAF() for variant "' + variant.id + '".')
                # Assert
                self.assertEqual(expected_records, observed_records)

    def testGetAltAD(self):
        for curr_dataset in sorted(self.freq_data):
            if not curr_dataset.startswith("wout_spl"):
                # Write test data
                content = self.freq_data[curr_dataset]
                with open(self.tmp_variants, "w") as FH_variants:
                    FH_variants.write(content)
                # Parse
                expected_records = list()
                observed_records = list()
                with VCFIO(self.tmp_variants) as FH_vcf:
                    for variant in FH_vcf:
                        if "skipAD" not in variant.info:
                            try:
                                for idx_spl, curr_spl in enumerate(FH_vcf.samples):
                                    expected_AD = self.freq_expected[variant.info["expModel"]]["AD"][curr_spl]
                                    expected_records.append("{} {}".format(variant.id, expected_AD))
                                    observed_records.append("{} {}".format(variant.id, variant.getAltAD(curr_spl)))
                            except Exception:
                                raise Exception('Error in TestVCFRecord.testGetAltAD() for variant "' + variant.id + '".')
                # Assert
                self.assertEqual(expected_records, observed_records)

    def testGetDP(self):
        for curr_dataset in sorted(self.freq_data):
            if not curr_dataset.startswith("wout_spl"):
                # Write test data
                content = self.freq_data[curr_dataset]
                with open(self.tmp_variants, "w") as FH_variants:
                    FH_variants.write(content)
                # Parse
                expected_records = list()
                observed_records = list()
                with VCFIO(self.tmp_variants) as FH_vcf:
                    for variant in FH_vcf:
                        if "skipDP" not in variant.info:
                            try:
                                for curr_spl in FH_vcf.samples:
                                    expected_DP = self.freq_expected[variant.info["expModel"]]["DP"][curr_spl]
                                    expected_records.append("{} {}".format(variant.id, expected_DP))
                                    observed_records.append("{} {}".format(variant.id, variant.getDP(curr_spl)))
                            except Exception:
                                raise Exception('Error in TestVCFRecord.testGetDP() for variant "' + variant.id + '".')
                # Assert
                self.assertEqual(expected_records, observed_records)

    def testGetPopAltAD(self):
        for curr_dataset in sorted(self.freq_data):
            # Write test data
            content = self.freq_data[curr_dataset]
            with open(self.tmp_variants, "w") as FH_variants:
                FH_variants.write(content)
            # Parse
            expected_records = list()
            observed_records = list()
            with VCFIO(self.tmp_variants) as FH_vcf:
                for variant in FH_vcf:
                    try:
                        expected_AD = self.freq_expected[variant.info["expModel"]]["AD"]["pop"]
                        expected_records.append("{} {}".format(variant.id, expected_AD))
                        observed_records.append("{} {}".format(variant.id, variant.getPopAltAD()))
                    except Exception:
                        raise Exception('Error in TestVCFRecord.testGetPopAltAD() for variant "' + variant.id + '".')
            # Assert
            self.assertEqual(expected_records, observed_records)

    def testGetPopAltAF(self):
        for curr_dataset in sorted(self.freq_data):
            # Write test data
            content = self.freq_data[curr_dataset]
            with open(self.tmp_variants, "w") as FH_variants:
                FH_variants.write(content)
            # Parse
            expected_records = list()
            observed_records = list()
            with VCFIO(self.tmp_variants) as FH_vcf:
                for variant in FH_vcf:
                    try:
                        expected_AF = self.freq_expected[variant.info["expModel"]]["AF"]["pop"]
                        expected_records.append("{} {}".format(variant.id, expected_AF))
                        observed_records.append("{} {}".format(variant.id, variant.getPopAltAF()))
                    except Exception:
                        raise Exception('Error in TestVCFRecord.testGetPopAltAF() for variant "' + variant.id + '".')
            # Assert
            self.assertEqual(expected_records, observed_records)

    def testGetPopDP(self):
        for curr_dataset in sorted(self.freq_data):
            # Write test data
            content = self.freq_data[curr_dataset]
            with open(self.tmp_variants, "w") as FH_variants:
                FH_variants.write(content)
            # Parse
            expected_records = list()
            observed_records = list()
            with VCFIO(self.tmp_variants) as FH_vcf:
                for variant in FH_vcf:
                    try:
                        expected_DP = self.freq_expected[variant.info["expModel"]]["DP"]["pop"]
                        expected_records.append("{} {}".format(variant.id, expected_DP))
                        observed_records.append("{} {}".format(variant.id, variant.getPopDP()))
                    except Exception:
                        raise Exception('Error in TestVCFRecord.testGetPopDP() for variant "' + variant.id + '".')
            # Assert
            self.assertEqual(expected_records, observed_records)

    def testNormalizeSingleAllele(self):
        # Test substitution one nt
        substitution = VCFRecord("artificial_1", 18, None, "TC", ["TA"], 230)
        substitution.normalizeSingleAllele()
        self.assertTrue(substitution.ref == "C" and substitution.alt[0] == "A" and substitution.pos == 19)
        # Test substitution multi nt
        substitution = VCFRecord("artificial_1", 18, None, "TCtgA", ["TaGCc"], 230)
        substitution.normalizeSingleAllele()
        self.assertTrue(substitution.ref == "CTGA" and substitution.alt[0] == "AGCC" and substitution.pos == 19)
        # Test substitution multi nt with possible remove
        substitution = VCFRecord("artificial_1", 18, None, "TCtgATT", ["TaGGctT"], 230)
        substitution.normalizeSingleAllele()
        self.assertTrue(substitution.ref == "CTGA" and substitution.alt[0] == "AGGC" and substitution.pos == 19)

        # Test insertion
        insertion = VCFRecord("artificial_1", 18, None, "T", ["TA"], 230)
        insertion.normalizeSingleAllele()
        self.assertTrue(insertion.ref == "-" and insertion.alt[0] == "A" and insertion.pos == 19)
        # Test insertion multi nt and remove upstream and downstream
        insertion = VCFRecord("artificial_1", 18, None, "TGAT", ["TCGAGAT"], 230)
        insertion.normalizeSingleAllele()
        self.assertTrue(insertion.ref == "-" and insertion.alt[0] == "CGA" and insertion.pos == 19)
        # Test insertion multi nt with possible remove and downstream and without complete standardization
        insertion = VCFRecord("artificial_1", 18, None, "TCtgATTAGC", ["TaGGctTATGCGC"], 230)
        insertion.normalizeSingleAllele()
        self.assertTrue(insertion.ref == "CTGATTA" and insertion.alt[0] == "AGGCTTATGC" and insertion.pos == 19)

        # Test deletion
        insertion = VCFRecord("artificial_1", 18, None, "TA", ["T"], 230)
        insertion.normalizeSingleAllele()
        self.assertTrue(insertion.ref == "A" and insertion.alt[0] == "-" and insertion.pos == 19)
        # Test insertion multi nt and remove upstream and downstream
        insertion = VCFRecord("artificial_1", 18, None, "TCGAGAT", ["TGAT"], 230)
        insertion.normalizeSingleAllele()
        self.assertTrue(insertion.ref == "CGA" and insertion.alt[0] == "-" and insertion.pos == 19)
        # Test insertion multi nt with possible remove and downstream and without complete standardization
        insertion = VCFRecord("artificial_1", 18, None, "TaGGctTATGCGC", ["TCtgATTAGC"], 230)
        insertion.normalizeSingleAllele()
        self.assertTrue(insertion.ref == "AGGCTTATGC" and insertion.alt[0] == "CTGATTA" and insertion.pos == 19)

    def testGetMostUpstream(self):
        ref = "nnNNATGCCAaTgATGTTtTaAGCCGAGCCGAT"  # length: 33
        #      | | | | | | | | | | | | | | | | |
        #      1 3 5 7 9 11| 15| 19| 23| 27| 31|
        #                  13  17  21  25  29  32
        # Test fix deletion
        deletion = VCFRecord("artificial_1", 18, None, "TtTaAGC", ["T"], 230)
        upstream = deletion.getMostUpstream(ref)
        self.assertTrue(upstream.pos == 19 and upstream.ref == "TTAAGC" and upstream.alt[0] == "-")
        # Test homopolymer deletion
        deletion = VCFRecord("artificial_1", 18, None, "TTT", ["T"], 230)
        upstream = deletion.getMostUpstream(ref)
        self.assertTrue(upstream.pos == 17 and upstream.ref == "TT" and upstream.alt[0] == "-")
        # Test complex deletion
        deletion = VCFRecord("artificial_1", 25, None, "CGAgCC", ["C"], 230)
        upstream = deletion.getMostUpstream(ref)
        self.assertTrue(upstream.pos == 22 and upstream.ref == "AGCCG" and upstream.alt[0] == "-")
        # Test deletion on start
        deletion = VCFRecord("artificial_1", 1, None, ref[0], ["-"], 230)
        upstream = deletion.getMostUpstream(ref)
        self.assertTrue(upstream.pos == 1 and upstream.ref == "N" and upstream.alt[0] == "-")
        # Test deletion on end
        deletion = VCFRecord("artificial_1", len(ref) - 1, None, ref[-2:], [ref[-2]], 230)
        upstream = deletion.getMostUpstream(ref)
        self.assertTrue(upstream.pos == 33 and upstream.ref == "T" and upstream.alt[0] == "-")
        # Test fix insertion
        insertion = VCFRecord("artificial_1", 4, None, "N", ["NGGTT"], 230)
        upstream = insertion.getMostUpstream(ref)
        self.assertTrue(upstream.pos == 5 and upstream.ref == "-" and upstream.alt[0] == "GGTT")
        # Test homopolymer insertion
        insertion = VCFRecord("artificial_1", 18, None, "T", ["TT"], 230)
        upstream = insertion.getMostUpstream(ref)
        self.assertTrue(upstream.pos == 17 and upstream.ref == "-" and upstream.alt[0] == "T")
        # Test complex insertion
        insertion = VCFRecord("artificial_1", 25, None, "C", ["CGAgCC"], 230)
        upstream = insertion.getMostUpstream(ref)
        self.assertTrue(upstream.pos == 22 and upstream.ref == "-" and upstream.alt[0] == "AGCCG")
        # Test insertion on start
        insertion = VCFRecord("artificial_1", 1, None, "-", ["A"], 230)
        upstream = insertion.getMostUpstream(ref)
        self.assertTrue(upstream.pos == 1 and upstream.ref == "-" and upstream.alt[0] == "A")
        # Test insertion on end
        insertion = VCFRecord("artificial_1", len(ref), None, ref[-1], [ref[-1] + "G"], 230)
        upstream = insertion.getMostUpstream(ref)
        self.assertTrue(upstream.pos == 34 and upstream.ref == "-" and upstream.alt[0] == "G")

    def testGetMostDownstream(self):
        ref = "nnNNATGCCAaTgATGTTtTaAGCCGAGCCGAT"  # length: 33
        #      | | | | | | | | | | | | | | | | |
        #      1 3 5 7 9 11| 15| 19| 23| 27| 31|
        #                  13  17  21  25  29  32
        # Test fix deletion
        deletion = VCFRecord("artificial_1", 18, None, "TtTaAGC", ["T"], 230)
        downstream = deletion.getMostDownstream(ref)
        self.assertTrue(downstream.pos == 19 and downstream.ref == "TTAAGC" and downstream.alt[0] == "-")
        # Test homopolymer deletion
        deletion = VCFRecord("artificial_1", 18, None, "TTT", ["T"], 230)
        downstream = deletion.getMostDownstream(ref)
        self.assertTrue(downstream.pos == 19 and downstream.ref == "TT" and downstream.alt[0] == "-")
        # Test complex deletion
        deletion = VCFRecord("artificial_1", 25, None, "CGAgCC", ["C"], 230)
        downstream = deletion.getMostDownstream(ref)
        self.assertTrue(downstream.pos == 28 and downstream.ref == "GCCGA" and downstream.alt[0] == "-")
        # Test deletion on start
        deletion = VCFRecord("artificial_1", 1, None, ref[0], ["-"], 230)
        downstream = deletion.getMostDownstream(ref)
        self.assertTrue(downstream.pos == 4 and downstream.ref == "N" and downstream.alt[0] == "-")
        # Test deletion on end
        deletion = VCFRecord("artificial_1", len(ref) - 1, None, ref[-2:], [ref[-2]], 230)
        downstream = deletion.getMostDownstream(ref)
        self.assertTrue(downstream.pos == 33 and downstream.ref == "T" and downstream.alt[0] == "-")
        # Test fix insertion
        insertion = VCFRecord("artificial_1", 4, None, "N", ["NGGTT"], 230)
        downstream = insertion.getMostDownstream(ref)
        self.assertTrue(downstream.pos == 5 and downstream.ref == "-" and downstream.alt[0] == "GGTT")
        # Test homopolymer insertion
        insertion = VCFRecord("artificial_1", 18, None, "T", ["TT"], 230)
        downstream = insertion.getMostDownstream(ref)
        self.assertTrue(downstream.pos == 21 and downstream.ref == "-" and downstream.alt[0] == "T")
        # Test complex insertion
        insertion = VCFRecord("artificial_1", 25, None, "C", ["CGAgCC"], 230)
        downstream = insertion.getMostDownstream(ref)
        self.assertTrue(downstream.pos == 33 and downstream.ref == "-" and downstream.alt[0] == "GCCGA")
        # Test insertion on start
        insertion = VCFRecord("artificial_1", 1, None, "-", ["A"], 230)
        downstream = insertion.getMostDownstream(ref)
        self.assertTrue(downstream.pos == 1 and downstream.ref == "-" and downstream.alt[0] == "A")
        # Test insertion on end
        insertion = VCFRecord("artificial_1", len(ref), None, ref[-1], [ref[-1] + "G"], 230)
        downstream = insertion.getMostDownstream(ref)
        self.assertTrue(downstream.pos == 34 and downstream.ref == "-" and downstream.alt[0] == "G")

    def testFastDownstreamed(self):
        ref = "nnNNATGCCAaTgATGTTtTaAGCCGAGCCGAT"  # length: 33
        #      | | | | | | | | | | | | | | | | |
        #      1 3 5 7 9 11| 15| 19| 23| 27| 31|
        #                  13  17  21  25  29  33
        with open(self.tmp_fa, "w") as writer:
            writer.write(">artificial_1\n")
            writer.write("nnNNATGCCAaTgATGTTtTaAGCCGAGCCGAT")
        with open(self.tmp_fai, "w") as writer:
            writer.write("artificial_1\t33\t14\t33\t34")
        with IdxFastaIO(self.tmp_fa) as seq_handler:
            # Test fix deletion
            deletion = VCFRecord("artificial_1", 18, None, "TTTaAGC", ["T"], 230)
            downstream = deletion.fastDownstreamed(seq_handler, 10)
            self.assertTrue(downstream.pos == 18 and downstream.ref == "TTTAAGC" and downstream.alt[0] == "T")
            # Test homopolymer deletion
            deletion = VCFRecord("artificial_1", 16, None, "GTT", ["G"], 230)
            downstream = deletion.fastDownstreamed(seq_handler, 10)
            self.assertTrue(downstream.pos == 18 and downstream.ref == "TTT" and downstream.alt[0] == "T")
            # Test complex deletion
            deletion = VCFRecord("artificial_1", 25, None, "CGAgCC", ["C"], 230)
            downstream = deletion.fastDownstreamed(seq_handler, 10)
            self.assertTrue(downstream.pos == 27 and downstream.ref == "AGCCGA" and downstream.alt[0] == "A")
            # Test deletion on start
            deletion = VCFRecord("artificial_1", 1, None, ref[0], ["-"], 230)
            downstream = deletion.fastDownstreamed(seq_handler, 10)
            self.assertTrue(downstream.pos == 3 and downstream.ref == "NN" and downstream.alt[0] == "N")
            # Test deletion on end
            deletion = VCFRecord("artificial_1", len(ref) - 1, None, ref[-2:], [ref[-2]], 230)
            downstream = deletion.fastDownstreamed(seq_handler, 10)
            self.assertTrue(downstream.pos == 32 and downstream.ref == "AT" and downstream.alt[0] == "A")
            # Test fix insertion
            insertion = VCFRecord("artificial_1", 4, None, "N", ["NGGTT"], 230)
            downstream = insertion.fastDownstreamed(seq_handler, 10)
            self.assertTrue(downstream.pos == 4 and downstream.ref == "N" and downstream.alt[0] == "NGGTT")
            # Test homopolymer insertion
            insertion = VCFRecord("artificial_1", 18, None, "T", ["TT"], 230)
            downstream = insertion.fastDownstreamed(seq_handler, 10)
            self.assertTrue(downstream.pos == 20 and downstream.ref == "T" and downstream.alt[0] == "TT")
            # Test complex insertion
            insertion = VCFRecord("artificial_1", 25, None, "C", ["CGAgCC"], 230)
            downstream = insertion.fastDownstreamed(seq_handler, 10)
            self.assertTrue(downstream.pos == 32 and downstream.ref == "A" and downstream.alt[0] == "AGCCGA")
            # Test insertion on end
            insertion = VCFRecord("artificial_1", len(ref), None, ref[-1], [ref[-1] + "G"], 230)
            downstream = insertion.fastDownstreamed(seq_handler, 10)
            self.assertTrue(downstream.pos == 33 and downstream.ref == "T" and downstream.alt[0] == "TG")

    def testRefStart(self):
        data = [
            # Substitutions
            ["artificial_1", 10, "substit_1", "A", ["T"], None, None, {"expected_start": 10, "expected_end": 10}],
            ["artificial_1", 10, "substit_2", "AA", ["TT"], None, None, {"expected_start": 10, "expected_end": 11}],
            # Deletions
            ["artificial_1", 12, "del_1", "A", ["-"], None, None, {"expected_start": 12, "expected_end": 12}],
            ["artificial_1", 10, "del_2", "AAA", ["-"], None, None, {"expected_start": 10, "expected_end": 12}],
            ["artificial_1", 10, "del_3", "AAA", ["A"], None, None, {"expected_start": 11, "expected_end": 12}],
            ["artificial_1", 10, "del_4", "AGC", ["A"], None, None, {"expected_start": 11, "expected_end": 12}],
            ["artificial_1", 10, "del_5", "AAAT", ["TG"], None, None, {"expected_start": 10, "expected_end": 13}],
            # Insertions
            ["artificial_1", 12, "ins_1", "A", ["TG"], None, None, {"expected_start": 12, "expected_end": 12}],
            ["artificial_1", 12, "ins_2", "A", ["AGT"], None, None, {"expected_start": 12.5, "expected_end": 12.5}],
            ["artificial_1", 11, "ins_3", "AA", ["AGT"], None, None, {"expected_start": 12, "expected_end": 12}],
            ["artificial_1", 10, "ins_4", "AAA", ["AAGT"], None, None, {"expected_start": 12, "expected_end": 12}],
            ["artificial_1", 10, "ins_5", "AATC", ["AACGT"], None, None, {"expected_start": 12, "expected_end": 13}],
        ]
        for curr_data in data:
            record = VCFRecord(*curr_data)
            self.assertTrue(record.refStart() == record.info["expected_start"])

    def testRefEnd(self):
        data = [
            # Substitutions
            ["artificial_1", 10, "substit_1", "A", ["T"], None, None, {"expected_start": 10, "expected_end": 10}],
            ["artificial_1", 10, "substit_2", "AA", ["TT"], None, None, {"expected_start": 10, "expected_end": 11}],
            # Deletions
            ["artificial_1", 12, "del_1", "A", ["-"], None, None, {"expected_start": 12, "expected_end": 12}],
            ["artificial_1", 10, "del_2", "AAA", ["-"], None, None, {"expected_start": 10, "expected_end": 12}],
            ["artificial_1", 10, "del_3", "AAA", ["A"], None, None, {"expected_start": 11, "expected_end": 12}],
            ["artificial_1", 10, "del_4", "AGC", ["A"], None, None, {"expected_start": 11, "expected_end": 12}],
            ["artificial_1", 10, "del_5", "AAAT", ["TG"], None, None, {"expected_start": 10, "expected_end": 13}],
            # Insertions
            ["artificial_1", 12, "ins_1", "A", ["TG"], None, None, {"expected_start": 12, "expected_end": 12}],
            ["artificial_1", 12, "ins_2", "A", ["AGT"], None, None, {"expected_start": 12.5, "expected_end": 12.5}],
            ["artificial_1", 11, "ins_3", "AA", ["AGT"], None, None, {"expected_start": 12, "expected_end": 12}],
            ["artificial_1", 10, "ins_4", "AAA", ["AAGT"], None, None, {"expected_start": 12, "expected_end": 12}],
            ["artificial_1", 10, "ins_5", "AATC", ["AACGT"], None, None, {"expected_start": 12, "expected_end": 13}],
        ]
        for curr_data in data:
            record = VCFRecord(*curr_data)
            self.assertTrue(record.refEnd() == record.info["expected_end"])

    def testIsDeletion(self):
        data = [
            {"variant": VCFRecord("chr1", 2, None, "A", ["T"]), "expected": False},
            {"variant": VCFRecord("chr1", 2, None, "A", ["AT"]), "expected": False},
            {"variant": VCFRecord("chr1", 1, None, "CA", ["C"]), "expected": True},
            {"variant": VCFRecord("chr1", 15, None, "CA", ["GT"]), "expected": False},
            {"variant": VCFRecord("chr1", 15, None, "CA", ["GTT"]), "expected": False},
            {"variant": VCFRecord("chr1", 15, None, "CAC", ["GT"]), "expected": True}
        ]
        for curr in data:
            self.assertEqual(curr["variant"].isDeletion(), curr["expected"])

    def testIsInsAndDel(self):
        data = [
            {"variant": VCFRecord("chr1", 2, None, "A", ["T"]), "expected": False},
            {"variant": VCFRecord("chr1", 2, None, "A", ["AT"]), "expected": False},
            {"variant": VCFRecord("chr1", 1, None, "CA", ["C"]), "expected": False},
            {"variant": VCFRecord("chr1", 15, None, "CA", ["GT"]), "expected": True},
            {"variant": VCFRecord("chr1", 15, None, "CA", ["GTT"]), "expected": True},
            {"variant": VCFRecord("chr1", 15, None, "CAC", ["GT"]), "expected": True}
        ]
        for curr in data:
            self.assertEqual(curr["variant"].isInsAndDel(), curr["expected"])

    def testIsInsertion(self):
        data = [
            {"variant": VCFRecord("chr1", 2, None, "A", ["T"]), "expected": False},
            {"variant": VCFRecord("chr1", 2, None, "A", ["AT"]), "expected": True},
            {"variant": VCFRecord("chr1", 1, None, "CA", ["C"]), "expected": False},
            {"variant": VCFRecord("chr1", 15, None, "CA", ["GT"]), "expected": False},
            {"variant": VCFRecord("chr1", 15, None, "CA", ["GTT"]), "expected": True},
            {"variant": VCFRecord("chr1", 15, None, "CAC", ["GT"]), "expected": False}
        ]
        for curr in data:
            self.assertEqual(curr["variant"].isInsertion(), curr["expected"])


class TestVCFHeader(unittest.TestCase):
    def setUp(self):
        tmp_folder = tempfile.gettempdir()
        unique_id = str(uuid.uuid1())

        # Temporary files
        self.tmp_in_variants = os.path.join(tmp_folder, unique_id + "_in.vcf")
        self.tmp_out_variants = os.path.join(tmp_folder, unique_id + "_out.vcf")
        content = """##fileformat=VCFv4.3
##fileDate=20090805
##source=myImputationProgramV3.1
##reference=file:///seq/references/1000GenomesPilot-NCBI36.fasta
##contig=<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,species="Homo sapiens",taxonomy=x>
##phasing=partial
##SAMPLE=<ID=NA00001,Assay="WholeGenome",Description="Patient germline genome from unaffected",Disease="None",DOI="url",Ethnicity="AFR">
##SAMPLE=<ID=NA00002,Assay="Exome",Description="European patient exome from breast cancer",Disease="Cancer",Ethnicity="CEU",Tissue="Breast">
##PEDIGREE=<ID=TumourSample,Original=GermlineID>
##PEDIGREE=<ID=SomaticNonTumour,Original=GermlineID>
##PEDIGREE=<ID=ChildID,Father=FatherID,Mother=MotherID>
##PEDIGREE=<ID=SampleID,Name_1=Ancestor_1,Name_N=Ancestor_N>
##INFO=<ID=NS,Number=1,Type=Integer,Description="Number of Samples With Data",Source="myImputationProgram",Version="3.1">
##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
##INFO=<ID=AF,Number=A,Type=Float,Description="Allele Frequency">
##INFO=<ID=SRC,Number=.,Type=String,Description="Caller source">
##INFO=<ID=AA,Number=1,Type=String,Description="Ancestral Allele">
##INFO=<ID=DB,Number=0,Type=Flag,Description="dbSNP membership",Version="129">
##INFO=<ID=H2,Number=0,Type=Flag,Description="HapMap2 membership">
##FILTER=<ID=q10,Description="Quality below 10">
##FILTER=<ID=s50,Description="Less than 50% of samples have data">
##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype \\"test\\"">
##FORMAT=<ID=GQ,Number=1,Type=Integer,Description="Genotype Quality">
##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read Depth">
##FORMAT=<ID=HQ,Number=2,Type=Integer,Description="Haplotype Quality">
#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT	NA00001	NA00002	NA00003
"""
        with open(self.tmp_in_variants, "w") as FH_variants:
            FH_variants.write(content)

    def tearDown(self):
        # Clean temporary files
        for curr_file in [self.tmp_in_variants, self.tmp_out_variants]:
            if os.path.exists(curr_file):
                os.remove(curr_file)

    def testReadWriteHeader(self):
        with VCFIO(self.tmp_in_variants) as FH_in:
            with VCFIO(self.tmp_out_variants, "w") as FH_out:
                FH_out.copyHeader(FH_in)
                FH_out.writeHeader()
        expected = None
        with open(self.tmp_in_variants) as FH_exp:
            expected = FH_exp.readlines()
        observed = None
        with open(self.tmp_out_variants) as FH_obs:
            observed = FH_obs.readlines()
        self.assertTrue(len(expected) != 0)
        self.assertEqual(sorted(expected), sorted(observed))

    def testHeaderAttrDel(self):
        data = HeaderInfoAttr(
            id="NS",
            number="1",
            type="Integer",
            description="Number of Samples With Data",
            source="myImputationProgram",
            version="3.1"
        )
        del(data.source)
        expected = {"id": "NS", "number": "1", "type": "Integer", "description": "Number of Samples With Data", "version": "3.1"}
        self.assertEqual(expected, data.datastore)

    def testHeaderAttrKeys(self):
        data = HeaderInfoAttr(
            id="NS",
            number="1",
            type="Integer",
            description="Number of Samples With Data",
            source="myImputationProgram",
            version="3.1"
        )
        expected = sorted(["id", "number", "type", "description", "source", "version"])
        self.assertEqual(expected, sorted(data.keys()))

    def testHeaderAttrStr(self):
        data = HeaderInfoAttr(
            id="NS",
            number="1",
            type="Integer",
            description="Number of Samples With Data",
            source="myImputationProgram",
            version="3.1"
        )
        expected = '<ID=NS,Number=1,Type=Integer,Description="Number of Samples With Data",Source="myImputationProgram",Version="3.1">'
        self.assertEqual(expected, str(data))

    def testGetHeaderAttr(self):
        data = [
            {
                "text": '##INFO=<ID=NS,Number=1,Type=Integer,Description="Number of Samples With Data",Source="myImputationProgram",Version="3.1">',
                "expected_dict": {"id": "NS", "number": "1", "type": "Integer", "description": "Number of Samples With Data", "source": "myImputationProgram", "version": "3.1"},
                "expected_cls": "HeaderInfoAttr"
            },
            {
                "text": '##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">',
                "expected_dict": {"id": "DP", "number": "1", "type": "Integer", "description": "Total Depth"},
                "expected_cls": "HeaderInfoAttr"
            },
            {
                "text": '##INFO=<ID=AF,Number=A,Type=Float,Description="Allele Frequency">',
                "expected_dict": {"id": "AF", "number": "A", "type": "Float", "description": "Allele Frequency"},
                "expected_cls": "HeaderInfoAttr"
            },
            {
                "text": '##INFO=<ID=SRC,Number=.,Type=String,Description="Caller source">',
                "expected_dict": {"id": "SRC", "number": ".", "type": "String", "description": "Caller source"},
                "expected_cls": "HeaderInfoAttr"
            },
            {
                "text": '##INFO=<ID=DB,Number=0,Type=Flag,Description="dbSNP membership, 2018-01",Version="129">',
                "expected_dict": {"id": "DB", "number": "0", "type": "Flag", "description": "dbSNP membership, 2018-01", "version": "129"},
                "expected_cls": "HeaderInfoAttr"
            },
            {
                "text": '##FILTER=<ID=q10,Description="Quality below 10",Source="myImputationProgram",Version=10>',
                "expected_dict": {"id": "q10", "description": "Quality below 10", "source": "myImputationProgram", "version": "10"},
                "expected_cls": "HeaderFilterAttr"
            },
            {
                "text": '##FILTER=<ID=s50,Description="Less than 50% of samples have data">',
                "expected_dict": {"id": "s50", "description": "Less than 50% of samples have data"},
                "expected_cls": "HeaderFilterAttr"
            },
            {
                "text": '##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">',
                "expected_dict": {"id": "GT", "number": "1", "type": "String", "description": "Genotype"},
                "expected_cls": "HeaderFormatAttr"
            },
            {
                "text": '##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype \\"test\\"">',
                "expected_dict": {"id": "GT", "number": "1", "type": "String", "description": 'Genotype "test"'},
                "expected_cls": "HeaderFormatAttr"
            }
        ]
        for curr in data:
            attr = getHeaderAttr(curr["text"])
            self.assertEqual(curr["expected_dict"], attr.datastore)
            self.assertEqual(curr["expected_cls"], attr.__class__.__name__)


########################################################################
#
# MAIN
#
########################################################################
if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""Classes and functions for reading/writing/processing files containing 2D matrix."""

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2017 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.0.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

import numpy as np


class DistanceMatrixIO:
    """Class to read and write a distance matrix from or to separate value file."""

    def __init__(self, path, mode="r", separator="\t", cast_fct=float):
        """
        Build and return an instance of DistanceMatrixIO.

        :param path: Path to the distances file.
        :type path: str
        :param mode: Mode to open the file ('r', 'w', 'a').
        :type mode: str
        :param separator: Separator used between values.
        :type separator: str
        :param cast_fct: Function used to cast values in matrix.
        :type cast_fct: fct
        :return: The new instance.
        :rtype: matrix.DistanceMatrixIO
        """
        self.cast_fct = cast_fct
        self.dist_matrix = list()
        self.filepath = path
        self.more = mode
        if mode != "r":
            raise NotImplementedError(
                "Only read mode is currently implemented for {}.".format(
                    self.__class__.__name__
                )
            )
        self.names = None
        self.separator = separator
        self._parse()

    def _parse(self):
        """
        Parse file to load data in instance.

        :todo: Select parser by input (_parseDense(1), _parseDense(0), _parseTriangular(), _parseSparse()).
        """
        self._parseDense()

    def _parseDense(self, col_offset=1):
        """
        Parse file in dense format to load data in instance.

        :param col_offset: Skip this number of columns from the left of file. This option is used to skip colnames.
        :param col_offset: int
        """
        with open(self.filepath) as FH_in:
            self.names = [elt.strip() for elt in FH_in.readline().split(self.separator)]
            self.names = self.names[1:]
            for row in FH_in:
                row_fields = list()
                for elt in row.split(self.separator)[col_offset:]:
                    elt = elt.strip()
                    if elt == "":
                        elt = np.nan
                    else:
                        elt = self.cast_fct(elt)
                    row_fields.append(elt)
                self.dist_matrix.append(row_fields)
        self.dist_matrix = np.array(self.dist_matrix)



# -*- coding: utf-8 -*-
"""Classes and functions for parsing and managing HGVS and mutalyzer results."""


__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2019 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.2.1'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

from anacore.sequence import AA3LettersAlphabet
import base64
from collections import deque
import re
import requests
import time
import urllib.parse


AA_ONE_BY_THREE = AA3LettersAlphabet.one_by_three

AA_THREE_BY_ONE = {value: key for key, value in AA_ONE_BY_THREE.items()}

ONE_LETTER_AA_LEXIC = set(AA_ONE_BY_THREE.values())

THREE_LETTERS_AA_LEXIC = AA3LettersAlphabet.words | {"*", "X"}


class Accession:
    """Class to manipulate accession number."""

    def __init__(self, id, version=None, source=None):
        """
        Build and return an instance of Accession.

        :param id: ID of the element (from NCBI accession: NP_001245144.1 => id.version).
        :type id: str
        :param version: Version of the element (from NCBI accession: NP_001245144.1 => id.version).
        :type version: int
        :param source: Source of the element. Example: NCBI or ENSEMBL
        :type source: str
        :return: The new instance.
        :rtype: Accession
        """
        self.id = id
        self.version = int(version) if version is not None else None
        self.source = source

    def __str__(self):
        """
        Return the “informal” or nicely printable string representation of the object.

        :return: The printable string representation of the object.
        :rtype: str
        """
        repr = self.id
        if self.version is not None:
            repr = "{}.{}".format(self.id, self.version)
        return repr


class HGVS:
    """Class to manipulate HGVS information."""

    def __init__(self, accession, type, change):
        """
        Build and return an instance of HGVS.

        :param accession: Accesion of the alterated element.
        :type accession: Accession
        :param type: HGVS type: "g" or "c" or "p".
        :type type: str
        :param change: Change on element alterated (from HGVS string: "accession:type.change").
        :type change: str
        :return: The new instance.
        :rtype: HGVS
        """
        self.accession = accession
        self.type = type
        self.change = change

    def isPredicted(self):
        """
        Return True if the change come from prediction (opposite to experimetal evidence).

        :return: True if the change come from prediction (opposite to experimetal evidence).
        :rtype: bool
        """
        return "(" in self.change or ")" in self.change

    def __str__(self):
        """
        Return the “informal” or nicely printable string representation of the object.

        :return: The printable string representation of the object.
        :rtype: str
        """
        return "{}:{}.{}".format(self.accession, self.type, self.change)

    @staticmethod
    def fromStr(hgvs_str):
        """
        Return an instance of HGVS from an HGVS string.

        :param hgvs_str: The HGVS string to parse.
        :type hgvs_str: str
        :return: The instance of HGVS.
        :rtype: HGVS
        """
        hgvs = None
        match = re.search(r"^(.+)\.(.+):(.)\.(.+)$", hgvs_str)
        if match is not None:
            acc_id, acc_version, hgvs_type, hgvs_change = match.groups()
            acc_obj = Accession(acc_id, acc_version)
            hgvs = HGVS(acc_obj, hgvs_type, hgvs_change)
        else:
            match = re.search(r"^(.+):(.)\.(.+)$", hgvs_str)
            if match is not None:
                acc_id, hgvs_type, hgvs_change = match.groups()
                acc_obj = Accession(acc_id)
                hgvs = HGVS(acc_obj, hgvs_type, hgvs_change)
            else:
                raise Exception('The HGVS "{}" has an invalid format.'.format(hgvs_str))
        return hgvs


class MutalyzerBatch:
    """
    Manage submission of query batch on a mutalyzer server.

    :Code example:
        .. highlight:: python
        .. code-block:: python

            mutalyzer_url = "https://mutalyzer.nl"
            queries = ["NC_000012.12:g.25245351C>T", "NG_012337.1(SDHD_v001):c.274G>T"]
            batch = MutalyzerBatch(mutalyzer_url, queries, "NameChecker")
            for curr_res in batch.syncRequest():
                print(res)
    """

    def __init__(self, queries, url="https://mutalyzer.nl", command="NameChecker", proxy_url=None):
        """
        Build and return an instance of MutalyzerBatch.

        :param url: URL for the mutalyzer server.
        :type url: str
        :param queries: List of queries (example: ["NC_000012.12:g.25245351C>T", "NG_012337.1(SDHD_v001):c.274G>T"]).
        :type queries: list
        :param command: Type of the batch job, choose from: NameChecker, SyntaxChecker, PositionConverter and SnpConverter.
        :type command: str
        :param proxy_url: URL for the proxy server if necessary.
        :type proxy_url: str
        :return: The new instance.
        :rtype: MutalyzerBatch
        """
        self._url = url
        self._command = command
        if command not in {"NameChecker", "SyntaxChecker", "PositionConverter", "SnpConverter"}:
            raise Exception("{} is not in authorized commands: {}".format(command, ["NameChecker", "SyntaxChecker", "PositionConverter", "SnpConverter"]))
        self._queries = queries
        self._proxy_url = proxy_url
        self._status = None
        self._progress = None
        self._id = None
        self._submit_time = None
        self._complete_time = None

    def getExecTime(self):
        """
        Return server execution time for the batch.

        :return: The server execution time for the batch.
        :rtype: float
        """
        return self._complete_time - self._submit_time

    def getRequestURL(self):
        """
        Return the request URL to submit the batch.

        :return: The request URL to submit the batch.
        :rtype: str
        """
        request_data = base64.urlsafe_b64encode("\n".join(self._queries).encode("utf-8"))
        request_data = urllib.parse.quote(request_data, safe='')
        url = '{}/json/submitBatchJob?process={};data={}'.format(self._url, self._command, request_data)
        return url

    def submit(self):
        """Submit the batch."""
        url_request = self.getRequestURL()
        self._submit_time = time.time()
        response = requests.get(
            url_request,
            proxies=(None if self._proxy_url is None else {"https": self._proxy_url, "http": self.self._proxy_url})
        )
        self._status = "submitted"
        if response.status_code != 200:
            self._status = "error"
            raise Exception("Request {} has failed.".format(url_request))
        self._id = response.json()

    def updateStatus(self):
        """Update progress and status of the submitted batch."""
        url_request = '{}/json/monitorBatchJob?job_id={}'.format(self._url, self._id)
        response = requests.get(
            url_request,
            proxies=(None if self._proxy_url is None else {"https": self._proxy_url, "http": self.self._proxy_url})
        )
        if response.status_code != 200:
            self._status = "error"
            raise Exception("Request {} has failed.".format(url_request))
        nb_left = response.json()
        self._progress = (len(self._queries) - nb_left) / len(self._queries)
        if nb_left == 0:
            self._status = "complete"
            self._complete_time = time.time()

    def getResponse(self):
        """
        Return sever result for the submitted batch.

        :return: The sever result for the submitted batch.
        :rtype: str
        """
        url_request = '{}/json/getBatchJob?job_id={}'.format(self._url, self._id)
        response = requests.get(
            url_request,
            proxies=(None if self._proxy_url is None else {"https": self._proxy_url, "http": self.self._proxy_url})
        )
        if response.status_code != 200:
            self._status = "error"
            raise Exception("Request {} has failed.".format(url_request))
        self._status = "success"
        return str(base64.b64decode(response.json()), "utf-8")

    def getParsedResponse(self, time_to_update=2):
        """
        Return sever result for the submitted batch in list format.

        :param time_to_update: Number of seconds to wait before retry to get response from server. After the status becomes "complete", the server take few seconds to update last results.
        :type time_to_update: int
        :return: The sever result for the submitted batch in list format.
        :rtype: list
        """
        response_lines = self.getResponse().strip().split("\n")
        while len(response_lines) - 1 < len(self._queries):  # After complete the result must be update on server
            time.sleep(time_to_update)
            response_lines = self.getResponse().strip().split("\n")
        titles = response_lines.pop(0).split("\t")
        results = []
        for curr_res in response_lines:
            curr_res = curr_res.split("\t")
            results.append({key: val for key, val in zip(titles, curr_res)})
        return results

    def syncRequest(self, time_to_update=3, first_time_to_update=1):
        """
        Return sever result for the batch in list format.

        :param time_to_update: Number of seconds to wait between each update of status from server.
        :type time_to_update: int
        :param first_time_to_update: Number of seconds to wait before the first status update after submit.
        :type first_time_to_update: int
        :return: The sever result for the batch in list format.
        :rtype: list
        """
        self.submit()
        time.sleep(first_time_to_update)
        self.updateStatus()
        while self._status not in {"error", "complete"}:
            time.sleep(time_to_update)
            self.updateStatus()
        return self.getParsedResponse()


class RunMutalyzerLegend:
    """Class to manipulate field legend from runMutalyzer[Light]."""

    def __init__(self, legend_data):
        """
        Build and return an instance of RunMutalyzerLegend.

        :param legend_data: Legend field from runMutalyzer[Light].
        :type legend_data: list
        :return: The new instance.
        :rtype: RunMutalyzerLegend
        """
        self.data = legend_data

    def getIdByName(self):
        """
        Return element accession by name. Example: {"KIT_v001": "NM_000222.2", "KIT_i001": "NP_000213.1"}.

        :return: element accession by name. Example: {"KIT_v001": "NM_000222.2", "KIT_i001": "NP_000213.1"}.
        :rtype: dict
        """
        id_by_name = {}
        for elt in self.data:
            if "id" in elt and "name" in elt:
                id_by_name[elt["name"]] = elt["id"]
        return id_by_name

    def getProtBytr(self):
        """
        Return protein accession by transcript accession.

        :return: Protein accession by transcript accession.
        :rtype: dict
        """
        prot_by_tr = {}
        id_by_name = self.getIdByName()
        for elt_name, elt_id in id_by_name.items():
            if "_i" in elt_name:
                prot_acc = elt_id
                tr_acc = id_by_name[elt_name.replace("_i", "_v")]
                prot_by_tr[tr_acc] = prot_acc
        return prot_by_tr


class RunMutalyzerDescription:
    """Class to manipulate field proteinDescriptions or transcriptDescriptions from runMutalyzer[Light]."""

    def __init__(self, description_data):
        """
        Build and return an instance of RunMutalyzerLegend.

        :param description_data: proteinDescriptions or transcriptDescriptions field from runMutalyzer[Light].
        :type description_data: list
        :return: The new instance.
        :rtype: RunMutalyzerLegend
        """
        self.data = description_data

    def getByAccession(self, id_by_name):
        """
        Return HGVS by RefSeq accession from proteinDescriptions or transcriptDescriptions of runMutalyzer[Light].

        :param id_by_name: Protein/Transcript RefSeq accession by name. Example: {"KIT_v001": "NM_000222.2", "KIT_i001": "NP_000213.1"}.
        :type id_by_name: dict
        :return: HGVS by RefSeq accession.
        :rtype: dict
        """
        HGVS_by_elt = {}
        for elt in self.data:
            match_renamed = re.search(r"(.+)\((.+)\):(.)\.(.+)", elt)
            if match_renamed:
                ref_acc, elt_name, hgvs_type, change = match_renamed.groups()
                elt_acc = elt_name
                if elt_name in id_by_name:
                    elt_acc = id_by_name[elt_name]
                if change == "?":
                    HGVS_by_elt[elt_acc] = ""
                else:
                    HGVS_by_elt[elt_acc] = "{}:{}.{}".format(elt_acc, hgvs_type, change)
            else:
                match_std = re.search(r"(.+):(.)\.(.+)", elt)
                if match_std:
                    elt_acc, hgvs_type, change = match_std.groups()
                    if change == "?":
                        HGVS_by_elt[elt_acc] = ""
                    else:
                        HGVS_by_elt[elt_acc] = "{}:{}.{}".format(elt_acc, hgvs_type, change)
                else:
                    raise Exception("The feature description {} cannot be parsed.".format(elt))
        return HGVS_by_elt


class HGVSProtChange:
    """Class to manipulate change part of proteic HGVS (ex: "Val600Glu")."""

    def __init__(self, start_aa, start_pos, end_aa, end_pos, new_aa, evt, new_elts, predicted=True):
        """
        Build and return an instance of HGVSProtChange.

        :param start_aa: First amino acid impacted by the variant. Example: "Val" in "Val600Glu" or "Gln" in "Gln746_Lys747ins*63" or "Ile" in "Ile327Argfs*?".
        :type start_aa: str
        :param start_pos: Proteic position of the first amino acid impacted by the variant. Example: 600 in "Val600Glu" or 746 in "Gln746_Lys747ins*63" or 327 in "Ile327Argfs*?".
        :type start_pos: int
        :param end_aa: Last amino acid impacted by the variant. Example: None in "Val600Glu" or "Lys" in "Gln746_Lys747ins*63" or None in "Ile327Argfs*?".
        :type end_aa: str
        :param end_pos: Proteic position of the last amino acid impacted by the variant. Example: None in "Val600Glu" or 747 in "Gln746_Lys747ins*63" or None in "Ile327Argfs*?".
        :type end_pos: int
        :param new_aa: Amino acid changed by "fs" or "ext". Example: None in "Val600Glu" or None in "Gln746_Lys747ins*63" or "Arg" in "Ile327Argfs*?".
        :type new_aa: str
        :param evt: Type of evenment in [None, "del", "dup", "ext", "fs", "ins", "insdel"]. Example: None in "Val600Glu" or "ins" in "Gln746_Lys747ins*63" or "fs" in "Ile327Argfs*?".
        :type evt: str
        :param new_elts: New elements list after evenment type in HGVS change. Example ["Glu"] from "Val600Glu" or ["*", "63"] from "Gln746_Lys747ins*63" or ["*", "?"] from "Ile327Argfs*?".
        :type new_elts: list
        :param predicted: True if HGVS consequence is predicted, i.e. without experimental evidence (no RNA or protein sequence analysed).
        :type predicted: bool
        :return: The new instance.
        :rtype: anacore.hgvs.HGVSProtChange
        """
        self.start_aa = start_aa
        self.start_pos = start_pos
        self.end_aa = end_aa
        self.end_pos = end_pos
        self.new_aa = new_aa
        self.evt = evt
        self.new_elts = [] if new_elts is None else new_elts
        self.predicted = predicted

    @property
    def new_aa(self):
        return self._new_aa

    @new_aa.setter
    def new_aa(self, new_val):
        self._new_aa = None if new_val is None else HGVSProtChange._cleanedAA(new_val)

    @property
    def start_aa(self):
        return self._start_aa

    @start_aa.setter
    def start_aa(self, new_val):
        self._start_aa = None if new_val is None else HGVSProtChange._cleanedAA(new_val)

    @property
    def end_aa(self):
        return self._end_aa

    @end_aa.setter
    def end_aa(self, new_val):
        self._end_aa = None if new_val is None else HGVSProtChange._cleanedAA(new_val)

    def __eq__(self, other):
        """
        Return true if the object represent the same HGVS change.

        :param other: HGVS change to compare.
        :type other: anacore.hgvs.HGVSProtChange
        :return: True if the object represent the same HGVS change.
        :rtype: bool
        """
        return repr(self) == repr(other)

    def __repr__(self):
        """
        Return the object representation in string.

        :return: The object representation in string.
        :rtype: str
        """
        rep_str = '"start_aa": {}, "start_pos": {}, ' \
            '"end_aa": {}, "end_pos": {}, ' \
            '"new_aa": {}, "evt": {}, ' \
            '"new_elts": {}, "predicted": {}'.format(
                self.start_aa,
                self.start_pos,
                self.end_aa,
                self.end_pos,
                self.new_aa,
                self.evt,
                self.new_elts,
                self.predicted
            )
        return '{' + rep_str + '}'

    def __str__(self):
        """
        Return the “informal” or nicely printable string representation of the object.

        :return: The printable string representation of the object.
        :rtype: str
        """
        if self.start_aa is None:
            if self.start_pos == 0:
                repr = "0"
            else:
                repr = "?"
        else:
            repr = "{}{}".format(self.start_aa, self.start_pos)
            if self.end_pos:
                repr += "_{}{}".format(self.end_aa, self.end_pos)
            if self.new_aa:
                repr += self.new_aa
            if self.evt:
                repr += self.evt
            repr += "".join(self.new_elts)
        if self.predicted:
            repr = "({})".format(repr)
        return repr

    def isInFrameIns(self):
        """
        Return True if the HGVS is insertion, duplication or repeat. Frameshift or extension resulting from insertion return False.

        :return: Return True if the HGVS is insertion, duplication or repeat.
        :rtype: bool
        """
        is_ins = False
        if self.evt in {"ins", "dup"}:
            is_ins = True
        elif self.evt is None and len(self.new_elts) != 0 and self.new_elts[0] == "[":  # Reapeat
            is_ins = True
        return is_ins

    @staticmethod
    def _cleanedAA(aa):
        """
        Validate and return an amino acid representation for HGVS.

        :type aa: Amino acid in one or three letter representation.
        :type aa: str
        :return: Return three letter representation.
        :rtype: str
        """
        aa = aa.capitalize()
        if len(aa) == 1:
            if aa not in ONE_LETTER_AA_LEXIC:
                raise Exception("{} is not a valid one letter amino acid. Authorized: {}.".change(aa, sorted(ONE_LETTER_AA_LEXIC)))
            aa = AA_THREE_BY_ONE[aa]
        elif len(aa) == 3:
            if aa not in THREE_LETTERS_AA_LEXIC:
                raise Exception("{} is not a valid three letter amino acid. Authorized: {}.".change(aa, sorted(THREE_LETTERS_AA_LEXIC)))
        else:
            raise Exception('The amino acid "{}" it is not in one or three letter representation.'.format(aa))
        return aa

    @staticmethod
    def _getAANbLetter(start_aa, follow):
        """
        Return the number of letters used to encode amino acids.

        :param start_aa: First amino acid impacted by the variant.
        :type start_aa: str
        :param follow: Part after first impacted positions in HGVS change (ex: "Glu" in Val600Glu or "_Lys747ins*63" in Gln746_Lys747ins*63 or "Argfs*?" in Ile327Argfs*?).
        :type follow: str
        :return: Number of letters used to encode amino acids.
        :rtype: int
        """
        aa_nb_letter = len(start_aa)
        if start_aa == "*":
            aa_nb_letter = None
            if follow.startswith("_"):
                idx = 1
                while not follow[idx].isdigit():
                    idx += 1
                aa_nb_letter = idx - 1
            elif len(follow) == 1 and follow not in {"?", "="}:
                aa_nb_letter = 1
            elif len(follow) >= 3:
                if follow[0:3] in THREE_LETTERS_AA_LEXIC:
                    aa_nb_letter = 3
                elif follow[0].isupper():
                    if follow[1:3] == "fs" and follow[0] in ONE_LETTER_AA_LEXIC:
                        aa_nb_letter = 1
                    elif len(follow) >= 4 and follow[1:4] == "ext" and follow[0] in ONE_LETTER_AA_LEXIC:
                        aa_nb_letter = 1
        return aa_nb_letter

    @staticmethod
    def _splittedOnEvt(change, aa_nb_letter, start_pos, end_pos, follow):
        """
        Return (new_aa, evenement, new_elts_string) from part after modified positions in HGVS change (ex: "Glu" in Val600Glu or "ins*63" in Gln746_Lys747ins*63 or "Argfs*?" in Ile327Argfs*?).

        :param change: Change part of HGVS string (ex: "Val600Glu").
        :type change: str
        :param aa_nb_letter: Number of letters to represent an amino acid: 1 or 3.
        :type aa_nb_letter: int
        :param start_pos: First amino acid changed by variant.
        :type start_pos: int
        :param end_pos: Last amino acid changed by variant.
        :type end_pos: int
        :param follow: Part after modified positions in HGVS change (ex: "Glu" in Val600Glu or "ins*63" in Gln746_Lys747ins*63 or "Argfs*?" in Ile327Argfs*?).
        :type follow: str
        :return: The tuple (new_aa, evenement, new_elts_string) from part after modified positions in HGVS change. Examples: (None, None, "Glu") from Glu or (None, "ins", "*63") from ins*63 or ("Arg", "fs", "*?") from Argfs*?.
        :rtype: (str, str, str)
        """
        curr_aa_lexic = THREE_LETTERS_AA_LEXIC if aa_nb_letter == 3 else ONE_LETTER_AA_LEXIC
        re_aa = "|".join(["({})".format(elt.lower()) for elt in curr_aa_lexic])
        re_aa = re_aa.replace("*", r"\*")
        new_aa = None
        evt = None
        lc_follow = follow.lower()
        if lc_follow.startswith("delins"):
            evt = "delins"
            follow = follow[6:]
            if len(follow) == 0:
                raise Exception('"delins" must be followed by the amino acids inserted: {}'.format(change))
        elif lc_follow.startswith("del"):
            evt = "del"
            follow = follow[3:]
            if len(follow) != 0:
                # Follow must be the deleted aa
                if len(follow) % aa_nb_letter != 0:
                    raise Exception('Deletion must be ended by "del" or the amino acids deleted: {}'.format(change))
                del_len = (start_pos if end_pos is None else end_pos) - start_pos + 1
                if len(follow) / aa_nb_letter != del_len:
                    raise Exception('Deletion must be ended by "del" or the amino acids deleted: {}'.format(change))
                del_aa = [follow[idx:idx+aa_nb_letter].capitalize() for idx in range(0, len(follow), aa_nb_letter)]
                for curr_del_aa in del_aa:
                    if curr_del_aa not in curr_aa_lexic:
                        raise Exception('Deletion must be ended by "del" or the amino acids deleted: {}'.format(change))
                follow = ""
        elif lc_follow.startswith("ins"):
            evt = "ins"
            follow = follow[3:]
            if len(follow) == 0:
                raise Exception("Insertion must be followed by the number or the amino acids inserted: {}".format(change))
            if end_pos - start_pos != 1:
                raise Exception("In insertion the end position must be the next amino acid after the start: {}".format(change))
        elif lc_follow.startswith("dup"):
            evt = "dup"
            follow = follow[3:]
            if len(follow) != 0:
                # Follow must be the deleted aa
                if len(follow) % aa_nb_letter != 0:
                    raise Exception('Duplication must be ended by "dup" or the amino acids duplicated: {}'.format(change))
                del_len = (start_pos if end_pos is None else end_pos) - start_pos + 1
                if len(follow) / aa_nb_letter != del_len:
                    raise Exception('Duplication must be ended by "dup" or the amino acids duplicated: {}'.format(change))
                del_aa = [follow[idx:idx+aa_nb_letter].capitalize() for idx in range(0, len(follow), aa_nb_letter)]
                for curr_del_aa in del_aa:
                    if curr_del_aa not in curr_aa_lexic:
                        raise Exception('Duplication must be ended by "dup" or the amino acids duplicated: {}'.format(change))
                follow = ""
        elif lc_follow == "ext" or lc_follow.startswith("ext-"):
            evt = "ext"
            follow = follow[3:]
        elif re.match(r"^(" + re_aa + ")ext", lc_follow):
            evt = "ext"
            new_aa = follow[:aa_nb_letter]
            follow = follow[aa_nb_letter + 3:]
        elif lc_follow == "fs":
            evt = "fs"
            follow = ""
        elif re.match(r"^(" + re_aa + ")fs", lc_follow):
            evt = "fs"
            new_aa = follow[:aa_nb_letter]
            follow = follow[aa_nb_letter + 2:]
        elif re.match(r"^fs(\*|ter)\d*$", lc_follow):
            evt = "fs"
            follow = follow[2:]
        return new_aa, evt, follow

    @staticmethod
    def _parsedNewElements(change, aa_nb_letter, evt, new_elements_str):
        """
        Return new elements list from part after evenment type in HGVS change (ex: "Glu" in Val600Glu or "*63" in Gln746_Lys747ins*63 or "*?" in Ile327Argfs*?).

        :param change: Change part of HGVS string (ex: "Val600Glu").
        :type change: str
        :param aa_nb_letter: Number of letters to represent an amino acid: 1 or 3.
        :type aa_nb_letter: int
        :param evt: Type of evenment in [None, "del", "dup", "ext", "fs", "ins", "insdel"].
        :type evt: str
        :param new_elements_str: Part after evenment type in HGVS change (ex: "Glu" in Val600Glu or "*63" in Gln746_Lys747ins*63 or "*?" in Ile327Argfs*?).
        :type new_elements_str: str
        :return: New elements list. Example ["Glu"] from "Glu" or ["*", "63"] from "*63" or ["*", "?"] from "*?".
        :rtype: list
        """
        new_elts = []
        remaining = deque(new_elements_str)
        curr_aa = ""
        while len(remaining):
            elt = remaining.popleft()
            if elt == "=":
                if len(remaining) > 0 and remaining[0] not in {"/", "^"}:
                    raise Exception('"=" it is not the last character in an alternative: {}'.format(change))
            elif elt == "*":
                elt = "Ter"
                if len(remaining) != 0:
                    if remaining[0] not in {"?", "^", "/"} and not remaining[0].isdigit():
                        raise Exception('"*" it is not the last element in change and the next element it is not in ["?", "/", "^", digit]: {}'.format(change))
            elif elt == "X":
                elt = "Xaa"
                if aa_nb_letter == 3:
                    if len(remaining) >= 2 and remaining[0] == "a" and remaining[1] == "a":
                        remaining.popleft()
                        remaining.popleft()
            elif elt in {"^", "/"}:
                if len(remaining) == 0:
                    raise Exception('"{}" cannot ends the change string: {}'.format(elt, change))
                ############################## TODO: parse evt in remaining: Val245=/del
            elif elt == "(":
                if evt != "ins":
                    raise Exception("Only insertions can contain brackets: {}.".format(change))
                while len(remaining) != 0 and remaining[0].isdigit():
                    elt += remaining.popleft()
                if elt == "(":
                    raise Exception("Only numbers are authorized in brackets: {}.".format(change))
                if len(remaining) == 0 or remaining[0] != ")":
                    raise Exception("Open bracket in insertion must be closed: {}.".format(change))
                elt += remaining.popleft()
            elif elt == "-":
                if evt != "ext":
                    raise Exception("Only extensions can contain minus: {}.".format(change))
                while len(remaining) != 0 and remaining[0].isdigit():
                    elt += remaining.popleft()
                if elt == "-":
                    raise Exception("Only numbers are authorized after minus: {}.".format(change))
            elif elt.isdigit():
                while len(remaining) != 0 and remaining[0].isdigit():
                    elt += remaining.popleft()
                if evt not in {"ins", "fs", "ext"}:
                    raise Exception('Digits are only authorized in change of type "ext", "fs" or "ins": {}.'.format(change))
                if evt != "ins":
                    if len(remaining) != 0 and remaining[0] != "/":
                        raise Exception('Digit are only authorized at the end of the change or followed by "/" for "fs" and "ext": {}'.format(change))
            elif elt == "?":
                if len(remaining) != 0:
                    raise Exception('"?" must be the last character in change: {}'.format(change))
            else:  # Is amino acid
                while len(elt) != aa_nb_letter:
                    elt += remaining.popleft()
                elt = elt.capitalize()
                if aa_nb_letter == 1:
                    elt = AA_THREE_BY_ONE[elt]
                else:
                    elt = elt.capitalize()
                    if elt not in THREE_LETTERS_AA_LEXIC:
                        raise Exception('The element "{}" in {} is not an amino acid.'.format(elt, change))
                if elt == "Ter":
                    if len(remaining) != 0:
                        if remaining[0] not in {"?", "^", "/"} and not remaining[0].isdigit():
                            raise Exception('"Ter" it is not the last alement in change and the next element it is not in ["?", "/", "^", digit]: {}'.format(change))
            new_elts.append(elt)
        if curr_aa != "":
            raise Exception("One of the amino acids is incomplete: {} in {}".format(curr_aa, change))
        return new_elts

    @staticmethod
    def isValid(change):
        """
        Return True if the string is a valid HGVS change (ex: "Val600Glu" or "(Val582_Asn583insXXXXX)").

        :param change: String to validate.
        :type change: str
        :return: True if the string is a valid HGVS change (ex: "Val600Glu" or "(Val582_Asn583insXXXXX)").
        :rtype: bool
        """
        is_hgvs = None
        try:
            HGVSProtChange.fromStr(change)
            is_hgvs = True
        except Exception:
            is_hgvs = False
        return is_hgvs

    @staticmethod
    def fromStr(change, aa_nb_letter=None):
        """
        Return HGVSProtChange from change part of HGVS string (ex: "Val600Glu").

        :param change: Change part of HGVS string (ex: "Val600Glu").
        :type change: str
        :param aa_nb_letter: Number of letters to represent an amino acid: 1 or 3. By default, the function tries to determine this number automatically.
        :type aa_nb_letter: int
        :return: HGVSProtChange from change part of HGVS string (ex: "Val600Glu").
        :rtype: anacore.hgvs.HGVSProtChange
        """
        predicted = None
        if change[0] == "(" and change[-1] == ")":
            change = change[1:-1]
            predicted = True
        else:
            predicted = False
        if change == "?":
            return HGVSProtChange(None, None, None, None, None, None, None, predicted)
        if change == "0":
            return HGVSProtChange(None, 0, None, None, None, None, None, predicted)
        match = re.search(r"^([^\d]+)(\d+)(.+)$", change)
        if not match:
            raise Exception('"{}" has not a valid syntax for HGVS protein change.'.format(change))
        else:
            start_aa, start_pos, follow = match.groups()
            start_pos = int(start_pos)
            start_aa.capitalize()
            if aa_nb_letter is None:
                aa_nb_letter = HGVSProtChange._getAANbLetter(start_aa, follow)
            end_aa = None
            end_pos = None
            if follow.startswith("_"):
                end_aa = follow[1:aa_nb_letter + 1]
                follow = follow[1 + aa_nb_letter:]
                end_pos = ""
                while follow[0].isdigit():
                    end_pos += follow[0]
                    follow = follow[1:]
                end_pos = int(end_pos)
            new_aa, evt, new_elts_str = HGVSProtChange._splittedOnEvt(change, aa_nb_letter, start_pos, end_pos, follow)
            new_elts = [elt for elt in new_elts_str]
            if len(new_elts_str) != 0:
                if not re.match(r"^\[\d+\]$", new_elts_str):  # The alteration is not a repeat
                    new_elts = HGVSProtChange._parsedNewElements(change, aa_nb_letter, evt, new_elts_str)
                else:  # The alteration is a repeat
                    nb_repeats = new_elts_str[1:-1]
                    if nb_repeats == "1":
                        new_elts = []
                        evt = "dup"
                    else:
                        new_elts = ["[", nb_repeats, "]"]
            return HGVSProtChange(start_aa, start_pos, end_aa, end_pos, new_aa, evt, new_elts, predicted)

    @staticmethod
    def insCouldBeIdentical(hgvs_ins, hgvs_repeat):
        """
        Return True if the insertion could be identical to the repeat (example: L16_I17insGTTL vs G13_L16dup). In this case, the difference bewtenn the two HGVS comes only from a difference of notation.

        :param hgvs_ins: HGVSp change for an insertion (example: L16_I17insGTTL).
        :type hgvs_ins: anacore.hgvs.HGVSProtChange
        :param hgvs_repeat: HGVSp change for a repeat (example: G13_L16dup or G13_L16[1]).
        :type hgvs_repeat: anacore.hgvs.HGVSProtChange
        :return: True if the insertion could be identical to the repeat.
        :rtype: bool
        """
        could_be_identical = False
        if hgvs_ins.isInFrameIns() and hgvs_repeat.isInFrameIns():
            repeat_end_pos = hgvs_repeat.end_pos if hgvs_repeat.end_pos is not None else hgvs_repeat.start_pos
            repeat_end_aa = hgvs_repeat.end_aa if hgvs_repeat.end_aa is not None else hgvs_repeat.start_aa
            len_dup = repeat_end_pos - hgvs_repeat.start_pos + 1
            nb_repeat = 1
            if hgvs_repeat.evt is None:  # Repeat
                nb_repeat = int(hgvs_repeat.new_elts[1])
            len_ins_in_repeat = len_dup * nb_repeat
            if len(hgvs_ins.new_elts) == len_ins_in_repeat:  # same length
                if repeat_end_aa + str(repeat_end_pos) == hgvs_ins.start_aa + str(hgvs_ins.start_pos):  # Same positions
                    if hgvs_repeat.start_aa == hgvs_ins.new_elts[0] and repeat_end_aa == hgvs_ins.new_elts[-1]:  # Same start and end amino acids
                        if nb_repeat == 1:
                            could_be_identical = True
                        else:
                            insertions = ["".join(hgvs_ins.new_elts[start:start + len_dup]) for start in range(0, len_ins_in_repeat, len_dup)]
                            if len(set(insertions)) == 1:  # Insertion is a valid repeat
                                could_be_identical = True
        return could_be_identical



#!/usr/bin/env python3

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2021 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.0.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

import os
import sys
import unittest

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
PACKAGE_DIR = os.path.dirname(TEST_DIR)
sys.path.append(PACKAGE_DIR)

from anacore.sequence import AA3LettersAlphabet, CodonAlphabet, DNAAlphabet, RNAAlphabet, Sequence


########################################################################
#
# FUNCTIONS
#
########################################################################
class TestAA3LettersAlphabet(unittest.TestCase):
    def testIsValid(self):
        self.assertTrue(AA3LettersAlphabet.isValid("AlaGlyHisTer"))
        self.assertFalse(AA3LettersAlphabet.isValid("AlaGlaHisTer"))  # Invalid aa
        self.assertFalse(AA3LettersAlphabet.isValid("AlaHiTer"))  # Invalid length

    def testToOneLetter(self):
        self.assertEqual(
            AA3LettersAlphabet.toOneLetter("AlaGlyHisTer"),
            "AGH*"
        )
        with self.assertRaises(Exception):
            AA3LettersAlphabet.toOneLetter("AlaGlaHisTer")  # Invalid aa
        with self.assertRaises(Exception):
            AA3LettersAlphabet.toOneLetter("AlaHiTer")  # Invalid length


class TestCodonAlphabet(unittest.TestCase):
    def testIsValid(self):
        self.assertTrue(CodonAlphabet.isValid("GCCGGGATCTGA"))
        self.assertFalse(CodonAlphabet.isValid("GCCGIGATCTGA"))  # Invalid nt
        self.assertFalse(CodonAlphabet.isValid("GCCGGGACTGA"))  # Invalid length

    def testTranslate(self):
        self.assertEqual(
            CodonAlphabet.translate("GCCGGGATCTGA"),
            "AlaGlyIleTer"
        )
        with self.assertRaises(Exception):
            CodonAlphabet.translate("GCCGIGATCTGA")  # Invalid nt
        with self.assertRaises(Exception):
            CodonAlphabet.translate("GCCGGGACTGA")  # Invalid length


class TestDNAAlphabet(unittest.TestCase):
    def testIsValid(self):
        self.assertTrue(DNAAlphabet.isValid("GCCGGGATCTGM"))
        self.assertFalse(DNAAlphabet.isValid("GCCGUGAICTGA"))  # Invalid nt

    def testRevCom(self):
        self.assertEqual(
            DNAAlphabet.revCom("GCCGGGATCTGM"),
            "KCAGATCCCGGC"
        )
        with self.assertRaises(Exception):
            DNAAlphabet.revCom("GCCGUGAICTGA")  # Invalid nt


class TestRNAAlphabet(unittest.TestCase):
    def testIsValid(self):
        self.assertTrue(RNAAlphabet.isValid("GCCGGGAUCUGM"))
        self.assertFalse(RNAAlphabet.isValid("GCCGIGAUCUGA"))  # Invalid nt

    def testRevCom(self):
        self.assertEqual(
            RNAAlphabet.revCom("GCCGGGAUCUGM"),
            "KCAGAUCCCGGC"
        )
        with self.assertRaises(Exception):
            RNAAlphabet.revCom("GCCGIGAUCUGA")  # Invalid nt


class TestSequence(unittest.TestCase):
    def testDnaRevCom(self):
        # Without quality
        test_record = Sequence("id1", "AtMGCUN").dnaRevCom()
        expected_revcom = Sequence("id1", "NAGCKaT")
        self.assertTrue(cmpSequences(test_record, expected_revcom))
        # With quality
        test_record = Sequence("id1", "AtMGCUN", "", "18AAEGH").dnaRevCom()
        expected_revcom = Sequence("id1", "NAGCKaT", "", "HGEAA81")
        self.assertTrue(cmpSequences(test_record, expected_revcom))

    def testRnaRevCom(self):
        # Without quality
        test_record = Sequence("id1", "AuMGCUN").rnaRevCom()
        expected_revcom = Sequence("id1", "NAGCKaU")
        self.assertTrue(cmpSequences(test_record, expected_revcom))
        # With quality
        test_record = Sequence("id1", "AuMGCUN", "", "18AAEGH").rnaRevCom()
        expected_revcom = Sequence("id1", "NAGCKaU", "", "HGEAA81")
        self.assertTrue(cmpSequences(test_record, expected_revcom))


def cmpSequences(rec_1, rec_2):
    are_equal = False
    if rec_1.id == rec_2.id:
        if rec_1.description == rec_2.description:
            if rec_1.string == rec_2.string:
                if rec_1.quality == rec_2.quality:
                    are_equal = True
    return are_equal


########################################################################
#
# MAIN
#
########################################################################
if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""Classes and functions for reading/writing/processing GFF files."""

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2017 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '2.0.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

import re
import copy
from anacore.abstractFile import AbstractFile, isEmpty
from anacore.region import Region


class GFF3Record(Region):
    """Class to manage a record comind from a GFF3."""

    def __init__(self, seq_id=None, source=None, type=None, start=None, end=None, score=None, strand=None, phase=None, attributes=None):
        """
        Build and return an instance of GFF3Record.

        :param seq_id: The ID of the landmark used to establish the coordinate system for the current feature.
        :type seq_id: str.
        :param source: The source is a free text qualifier intended to describe the algorithm or operating procedure that generated this feature. Typically this is the name of a piece of software, such as "Genescan" or a database name, such as "Genbank." In effect, the source is used to extend the feature ontology by adding a qualifier to the type creating a new composite type that is a subclass of the type in the type column.
        :type source: str.
        :param type: The type of the feature (previously called the "method"). This is constrained to be either: (a) a term from the "lite" sequence ontology, SOFA; or (b) a SOFA accession number. The latter alternative is distinguished using the syntax SO:000000.
        :type type: str.
        :param start: The start of the feature, in 1-based integer coordinates, relative to the landmark given in seq_id. Start is always less than or equal to end.
        :type start: int.
        :param end: The end of the feature, in 1-based integer coordinates, relative to the landmark given in seq_id. Start is always less than or equal to end.
        :type end: int.
        :param score: The score of the feature. As in earlier versions of the format, the semantics of the score are ill-defined. It is strongly recommended that E-values be used for sequence similarity features, and that P-values be used for ab initio gene prediction features.
        :type score: float.
        :param strand: The strand of the instance ("+" or "-").
        :type strand: str.
        :param phase: For features of type "CDS", the phase indicates where the feature begins with reference to the reading frame. The phase is one of the integers 0, 1, or 2, indicating the number of bases that should be removed from the beginning of this feature to reach the first base of the next codon. In other words, a phase of "0" indicates that the next codon begins at the first base of the region described by the current line, a phase of "1" indicates that the next codon begins at the second base of this region, and a phase of "2" indicates that the codon begins at the third base of this region. This is NOT to be confused with the frame, which is simply start modulo 3. For forward strand features, phase is counted from the start field. For reverse strand features, phase is counted from the end field.
        :type phase: int.
        :param attributes: The annotations of the feature.
        :type attributes: dict.
        :return: The new instance.
        :rtype: GFF3Record
        """
        name = None
        cleaned_attributes = attributes
        if attributes is not None and "Name" in attributes:
            name = attributes["Name"]
            cleaned_attributes = copy.deepcopy(attributes)
            del(cleaned_attributes["Name"])
        Region.__init__(self, start, end, strand, seq_id, name, cleaned_attributes)
        self.source = source
        self.type = type
        self.score = score
        self.phase = phase

    def __getattribute__(self, attr_name):
        """
        Return the value of the selected attribute.

        :param attr_name: The name of the selected attribute.
        :type attr_name: str
        :return: The value of the attribute.
        :rtype: *
        """
        attr_value = None
        if attr_name == "seq_id":
            attr_value = object.__getattribute__(self, "reference").name
        else:
            attr_value = object.__getattribute__(self, attr_name)
        return attr_value

    def __setattr__(self, attr_name, attr_value):
        """
        Set the value of the selected attribute.

        :param attr_name: The name of the selected attribute.
        :type attr_name: str
        :param attr_value: The new value for the selected attribute.
        :type attr_value: str
        """
        if attr_name == "seq_id":
            self.setReference(attr_value)
        else:
            object.__setattr__(self, attr_name, attr_value)

    def addToAnnot(self, tag, value):
        """
        Add one value on an existing tag.

        :param tag: tag of the annottation.
        :type tag: str.
        :param value: value to add.
        :type value: str.
        """
        if tag in self.annot:
            self.annot[tag] = self.annot[tag] + "," + value
        else:
            self.annot[tag] = value

    def _annotToGff(self):
        """
        Return a string in GFF3 field format corresponding to the attributes of the record.

        :return: The record.
        :rtype: str
        """
        gff_str = ""
        for tag, value in sorted(self.annot.items()):
            gff_str += "{}={};".format(GFF3IO.encodedValue(tag), GFF3IO.encodedValue(value))
        if gff_str != "":
            gff_str = gff_str[:-1]
        return gff_str

    def toGff(self):
        """
        Return the string corrsponding to the GFF3Record in GFF3 format.

        :return: Line in GFF3 format.
        :rtype: str
        """
        attributes = self._annotToGff()
        if self.name is not None:
            name = "{}={}".format("Name", GFF3IO.encodedValue(self.name))
            attributes = name + ";" + attributes
        return "{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}".format(
            self.seq_id,
            ("." if self.source is None else self.source),
            self.type,
            self.start,
            self.end,
            ("." if self.score is None else self.score),
            ("." if self.strand is None else self.strand),
            ("." if self.phase is None else self.phase),
            attributes
        )


class GFF3IO(AbstractFile):
    """Class to manage read and write in GFF3 file."""

    def __init__(self, filepath, mode="r", separator="\t", title_starter="#", has_title=True):
        """
        Build and return an instance of GFF3IO.

        :param filepath: The filepath.
        :type filepath: str.
        :param mode: Mode to open the file ('r', 'w', 'a').
        :type mode: str.
        :return: The new instance.
        :rtype: GFF3IO
        """
        # Convert mode for append in empty file
        if mode == "a" and isEmpty(filepath):
            mode = "w"
        # Initialise instance
        AbstractFile.__init__(self, filepath, mode)
        # Add header
        if mode == "w":
            self.file_handle.write("##gff-version 3\n")
            self.current_line_nb += 1

    def isRecordLine(self, line):
        """
        Return True if the line corresponds to a record (it is not a comment or an header line).

        :param line: The evaluated line.
        :type line: str
        :return: True if the line corresponds to a record.
        :rtype: bool
        """
        is_record = True
        if line.startswith("#"):
            is_record = False
        return is_record

    def _parseLine(self):
        """
        Return a structured record from the GFF current line.

        :return: The record.
        :rtype: GFF3Record
        """
        fields = [elt.strip() for elt in self.current_line.split("\t")]
        gff_record = GFF3Record(
            fields[0],
            (None if fields[1] == "." else fields[1]),
            fields[2],
            (None if fields[3] == "." else int(fields[3])),
            (None if fields[4] == "." else int(fields[4])),
            (None if fields[5] == "." else fields[5]),
            (None if fields[6] == "." else fields[6]),
            (None if fields[7] == "." else fields[7]),
        )
        # Parse attributes
        attributes_field = "\t".join(fields[8:])
        if attributes_field.strip().endswith(";"):  # if attributes end with ';'
            attributes_field = attributes_field.strip()[:-1]
        attributes_array = attributes_field.split(";")
        cleaned_attributes = list()
        for attribute in attributes_array:
            if "=" not in attribute:
                cleaned_attributes[len(cleaned_attributes) - 1] += ";" + attribute
            else:
                cleaned_attributes.append(attribute)
        for attribute in cleaned_attributes:
            matches = re.match("^([^=]+)=(.*)", attribute)
            tag = GFF3IO.decodedValue(matches.group(1).strip())
            values = [GFF3IO.decodedValue(elt) for elt in matches.group(2).split(',')]
            for current_val in values:
                if tag == "Name":
                    gff_record.name = current_val
                else:
                    gff_record.addToAnnot(tag, current_val)
        return gff_record

    def write(self, record):
        """
        Write one line on GFF file.

        :param record: The object to write.
        :type record: GFF3Record
        """
        self.file_handle.write(record.toGff() + "\n")
        self.current_line_nb += 1

    @staticmethod
    def decodedValue(encoded_value):
        """
        Return the attribute tag/value in human readable format.

        :param encoded_value: The attribute tag or value to decode.
        :type encoded_value: str
        :return: Human readable value.
        :rtype: str
        :see: RFC 3986 Percent-Encoding.
        """
        decoded_value = encoded_value.replace('%3B', ';').replace('%2C', ',').replace('%3D', '=')
        return decoded_value

    @staticmethod
    def encodedValue(dirty_value):
        """
        Return value after GFF3 attribute cleaning. Cleanning:

            - URL escaping rules are used for tags or values containing the following characters: ",=;".
            - Spaces are allowed in this field, but tabs must be replaced by space.
            - Quotes ' and " are deleted.

        :param dirty_value: The attribute tag or value to encode.
        :type dirty_value: str
        :return: GFF3 encoded value.
        :rtype: str
        :see: RFC 3986 Percent-Encoding.
        """
        cleaned_value = dirty_value.replace(';', '%3B')
        cleaned_value = cleaned_value.replace(',', '%2C')
        cleaned_value = cleaned_value.replace('=', '%3D')
        cleaned_value = cleaned_value.replace('\t', ' ')
        cleaned_value = cleaned_value.replace("'", '')
        cleaned_value = cleaned_value.replace('"', '')
        return cleaned_value



# -*- coding: utf-8 -*-
"""Classes and functions for reading Illumina's files (samplesheet, runInfo, runParameters, ...) and parsing header, filenames and run folder informations."""

__author__ = 'Frederic Escudie'
__copyright__ = 'Copyright (C) 2017 IUCT-O'
__license__ = 'GNU General Public License'
__version__ = '1.21.0'
__email__ = 'escudie.frederic@iuct-oncopole.fr'
__status__ = 'prod'

import os
import re
import glob
import datetime
import xml.etree.ElementTree as ET


class SampleSheetIO(object):
    """Reader for SampleSheet.csv."""

    SECTIONS = ["Header", "Manifests", "Reads", "Settings", "Data"]

    def __init__(self, path):
        self.filepath = path
        self.run = None
        self.samples = None
        self.header = None
        self.manifests = None
        self._parse()

    def _parse(self):
        # Retrieve information by section
        with open(self.filepath) as FH_sheet:
            sections_by_title = dict()
            section_title = None
            for line in FH_sheet:
                striped_line = line.strip()
                if any([re.fullmatch("\[" + elt + "\],*", striped_line) for elt in SampleSheetIO.SECTIONS]):
                    while striped_line.endswith(","):  # Removes invalid comma used in CSV to obtain the same number of columns in whole file
                        striped_line = striped_line[:-1]
                    section_title = striped_line[1:-1]
                    sections_by_title[section_title] = list()
                elif re.fullmatch(",*", striped_line):
                    section_title = None
                else:
                    sections_by_title[section_title].append(striped_line)
        # Process information
        self.samples = self._getSamplesFromData(sections_by_title["Data"])
        self.header = self._getInfoFromSection(sections_by_title["Header"])
        self.manifests = self._getInfoFromSection(sections_by_title["Manifests"]) if "Manifests" in sections_by_title else dict()
        self.run = self._getRunFromHeaderAndReads(sections_by_title["Header"], sections_by_title["Reads"])
        # Post process
        if "Description" not in self.header:
            self.header["Description"] = ""

    def _getSamplesFromData(self, data_section):
        samples = list()
        data_titles = [field.strip() for field in data_section[0].split(",")]
        for spl_idx, line in enumerate(data_section[1:]):
            spl = {data_titles[idx]: field.strip() for idx, field in enumerate(line.split(","))}
            spl["Sample_Basename"] = getIlluminaName(spl["Sample_ID"])
            spl["Library_Basename"] = spl["Sample_Basename"] + "_S" + str(spl_idx + 1)
            if "Sample_Description" in spl:
                spl["Description"] = spl["Sample_Description"]
            samples.append(spl)
        return samples

    def _getInfoFromSection(self, section):
        info = dict()
        for line in section:
            key, value = [field.strip() for field in line.split(",", 1)]
            while value.endswith(","):  # Removes invalid comma used in CSV to obtain the same number of columns in whole file
                value = value[:-1]
            info[key] = value
        return info

    def _getRunFromHeaderAndReads(self, header_section, reads_section):
        run = dict()
        # Header
        for line in header_section:
            key, value = line.split(",", 1)
            while value.endswith(","):  # Removes invalid comma used in CSV to obtain the same number of columns in whole file
                value = value[:-1]
            run[key] = value
        # Reads
        read_idx = 1
        for line in reads_section:
            striped_line = line.strip()
            while striped_line.endswith(","):  # Removes invalid comma used in CSV to obtain the same number of columns in whole file
                striped_line = striped_line[:-1]
            run["nb_cycles_R" + str(read_idx)] = int(striped_line)
            read_idx += 1
        return run


class ADSSampleSheetIO(SampleSheetIO):
    """Reader for SampleSheet.csv designed for AmpliconDS analysis."""

    def _getSamplesFromData(self, data_section):
        samples = super()._getSamplesFromData(data_section)
        for spl_idx, spl in enumerate(samples):
            spl["Sample_Basename"] = getIlluminaName(spl["Sample_Name"])
            spl["Library_Basename"] = spl["Sample_Basename"] + "_S" + str(spl_idx + 1)
        return samples

    def filterPanels(self, selected_manifests):
        """
        Filter samples and manifests on their panel.

        :param selected_manifests: The manifests corresponding to the selected panel.
        :type selected_manifests: list
        """
        # Get kept
        new_manifests = {}
        new_by_old_id = {}
        new_idx = 0
        for manifest_id in sorted(self.manifests):
            manifest_file = self.manifests[manifest_id]
            if manifest_file in selected_manifests:
                new_manifest_id = chr(ord("A") + new_idx)
                new_by_old_id[manifest_id] = new_manifest_id
                new_manifests[new_manifest_id] = self.manifests[manifest_id]
                new_idx += 1
        # Update manifests
        self.manifests = new_manifests
        # Update samples
        new_spl = list()
        for spl in self.samples:
            if spl["Manifest"] in new_by_old_id:
                if spl["Sample_ID"].endswith("_" + spl["Manifest"]):
                    spl["Sample_ID"] = spl["Sample_ID"].rsplit("_" + spl["Manifest"], 1)[0]
                    spl["Sample_ID"] += "_" + new_by_old_id[spl["Manifest"]]
                spl["Manifest"] = new_by_old_id[spl["Manifest"]]
                new_spl.append(spl)
        self.samples = new_spl

    def findSplFiles(self, directory, end_pattern, subject="library"):
        """
        Return by sample name or library name all the files in directory corresponding to the subject.

        :param directory: The path to the directory where the files names are evaluated.
        :type directory: str
        :param end_pattern: File pattern added at the end of each sample basename or library basename to select files.
        :type end_pattern: str
        :param subject: "library" or "sample" to select files corresponding to libraries or corresponding to samples.
        :type subject: str
        :return: By subject the paths of the corresponding files.
        :rtype: dict
        """
        subject_start_tag = subject.capitalize()
        files_by_elt = {}
        for spl in self.samples:
            if subject == "library" or spl["Sample_Name"] not in files_by_elt:  # The subject is the library or the subject is the sample and no library has been already processed for this sample
                files_by_elt[spl[subject_start_tag + "_Name"]] = sorted(glob.glob(
                    os.path.join(directory, spl[subject_start_tag + "_Basename"] + end_pattern)
                ))
        return files_by_elt

    def setSplFiles(self, tag, directory, end_pattern, subject="library"):
        """
        Add the files in directory corresponding to the subject. These information are keyed by provided tag for each element of self.samples.

        :param tag: The attribute used in self.sample to store the files paths.
        :type tag: str
        :param directory: The path to the directory where the files names are evaluated.
        :type directory: str
        :param end_pattern: File pattern added at the end of each sample basename or library basename to select files.
        :type end_pattern: str
        :param subject: "library" or "sample" to select files corresponding to libraries or corresponding to samples.
        :type subject: str
        """
        subject_tag = subject.capitalize() + "_Name"
        files_by_spl = self.findSplFiles(directory, end_pattern, subject)
        for spl in self.samples:
            spl[tag] = files_by_spl[spl[subject_tag]]


class Bcl2fastqLog(object):
    """Reader for bcl2fastq log."""

    def __init__(self, path):
        self.command = None
        self.complete = None
        self.end_time = None
        self.filepath = path
        self.start_time = None
        self.version = None
        self._parse()

    def _parse(self):
        self.command = None
        self.complete = None
        self.end_time = None
        self.start_time = None
        self.version = None
        last_info = ""
        with open(self.filepath) as reader:
            for line in reader:
                line = line.strip()
                if self.version is None and line.startswith("bcl2fastq v"):
                    self.version = line.split(" ")[1]
                elif self.command is None and "Command-line invocation:" in line:
                    trace_time, command = line.split(" Command-line invocation: ")
                    self.command = command
                    self.start_time = datetime.datetime.strptime(
                        trace_time.rsplit(" ", 1)[0],
                        '%Y-%m-%d %H:%M:%S'
                    )
                elif line != "":
                    last_info = line
        if "Processing completed" in last_info:
            self.end_time = datetime.datetime.strptime(
                " ".join(last_info.split(" ", 3)[:2]),
                '%Y-%m-%d %H:%M:%S'
            )
            self.complete = {
                "nb_errors": int(
                    re.search(r"(\d+) errors", last_info).group(1)
                ),
                "nb_warnings": int(
                    re.search(r"(\d+) warnings", last_info).group(1)
                )
            }


class RTAComplete(object):
    """Reader for RTAComplete.txt."""

    def __init__(self, path):
        self.filepath = path
        self.end_date = None
        self.RTA_version = None
        self._parse()

    def _parse(self):
        with open(self.filepath) as FH_in:
            content = FH_in.read().strip()
            match = re.fullmatch("(.+/.+/.+,.+),Illumina RTA (.+)", content)
            if match:  # 11/1/2017,15:11:43.174,Illumina RTA 1.18.54
                date_str, self.RTA_version = match.groups()
                if "." in date_str and ".":
                    date_str = date_str.split(".")[0]
                self.end_date = datetime.datetime.strptime(date_str, '%m/%d/%Y,%H:%M:%S')
            else:
                match = re.fullmatch("RTA (.+) completed on (.+/.+/.+ .+:.+:.+ ..)", content)
                if match:  # RTA 2.4.11 completed on 11/14/2019 4:56:45 AM
                    self.RTA_version, date_str = match.groups()
                    self.end_date = datetime.datetime.strptime(date_str, '%m/%d/%Y %I:%M:%S %p')
                else:
                    raise Exception('"{}" in {} cannot be parsed by {}.'.format(content, self.filepath, self.__class__.__name__))


class RunInfo(object):
    """Reader for RunInfo.xml."""

    def __init__(self, path):
        self.filepath = path
        self.flowcell = None  # {'id': 'HN33VBGX5', 'layout': {'LaneCount': '4', 'SurfaceCount': '2', 'SwathCount': '3', 'TileCount': '12', 'SectionPerLane': '3', 'LanePerSection': '2'}}
        self.instrument = None  # {'id': 'NS500523', 'platform': 'NextSeq'}
        self.reads_phases = None  # [{'is_index': False, 'nb_cycles': 151}, {'is_index': True, 'nb_cycles': 8}, {'is_index': True, 'nb_cycles': 8}, {'is_index': False, 'nb_cycles': 151}]
        self.run = None  # {'number': '133', 'id': '180406_NS500523_0133_AHN33VBGX5', 'start_date': datetime.datetime(2018, 4, 6, 0, 0)}
        self._parse()

    def _parse(self):
        # Retrieve information by section
        tree = ET.parse(self.filepath)
        run = tree.getroot().find("Run")
        # Process information
        self.instrument = self._getInstrumentFromRun(run)
        self.reads_phases = self._getReadsFromRun(run.find("Reads"))
        self.run = self._getRunFromRun(run)
        self.flowcell = self._getFlowcellFromRun(run)

    def _getReadsFromRun(self, reads_tree):
        reads = list()
        for child in reads_tree:
            reads.append({
                "is_index": (child.get("IsIndexedRead") == "Y"),
                "nb_cycles": int(child.get("NumCycles"))
            })
        return reads

    def _getInstrumentFromRun(self, run_tree):
        serial_number = run_tree.find("Instrument").text
        return {
            "id": serial_number,
            "platform": platformFromInstrumentSerialNumber(serial_number)
        }

    def _getRunFromRun(self, run_tree):
        date_str = run_tree.find("Date").text
        start_date = None
        if len(date_str) == 6:
            start_date = datetime.datetime.strptime(date_str, '%y%m%d')
        elif len(date_str) == 8:
            start_date = datetime.datetime.strptime(date_str, '%Y%m%d')
        else:
            start_date = datetime.datetime.strptime(date_str, '%m/%d/%Y %I:%M:%S %p')
        return {
            "number": run_tree.attrib["Number"],
            "id": run_tree.attrib["Id"],
            "start_date": start_date
        }

    def _getFlowcellFromRun(self, run_tree):
        return {
            "id": run_tree.find("Flowcell").text,
            "layout": run_tree.find("FlowcellLayout").attrib
        }


class RunParameters(object):
    """Reader for runParameters.xml."""

    def __init__(self, path):
        self.filepath = path
        self.kit = None  # {"flowcell_id": "H14U5BGXX", "reagent_kit_id": "NS2044178-REAGT"}
        self.instrument = None  # {"id": "NS500413", "platform": "NextSeq"}
        self.reads_phases = None  # [{'is_index': False, 'nb_cycles': 151}, {'is_index': True, 'nb_cycles': 8}, {'is_index': True, 'nb_cycles': 8}, {'is_index': False, 'nb_cycles': 151}]
        self.run = None  # {"number": "0033", "id": "141107_NS500413_0033_H14U5BGXX", "start_date": datetime}
        self.post_process = None  # "GenerateFASTQ"
        self.software = None  # {"RTA": "2.11.3", "CS": "4.0.1.41"}
        self._parse()

    def _parse(self):
        # Retrieve information by section
        tree = ET.parse(self.filepath)
        root = tree.getroot()
        if {"Version", "Setup"} == {child.tag for child in root}:  # For HiSeq information are all in <Setup> under <RunParameters>
            root = [child for child in root if child.tag == "Setup"][0]
        # Process information
        self.instrument = self._getInstrumentFromRoot(root)
        reads_subtree = root.find("Reads")
        if reads_subtree is not None:
            self.reads_phases = self._getReadsFromSection(reads_subtree)
        else:
            self.reads_phases = self._getReadsFromSetup(root.find("Setup"))
        self.run = self._getRunFromRoot(root)
        self.kit = self._getKitFromRoot(root)
        self.post_process = self._getPostProcessFromRoot(root)
        self.software = self._getSoftwareFromRoot(root)

    def _getReadsFromSetup(self, subtree):
        reads = list()
        indices = list()
        for child in subtree:
            if child.tag.startswith("Read"):
                reads.append({
                    "is_index": False,
                    "nb_cycles": int(child.text)
                })
            elif child.tag.startswith("Index"):
                indices.append({
                    "is_index": True,
                    "nb_cycles": int(child.text)
                })
        phases = [reads[0]]
        for curr in indices:
            phases.append(curr)
        for curr in reads[1:]:
            phases.append(curr)
        return phases

    def _getReadsFromSection(self, subtree):
        reads = list()
        for child in subtree:
            reads.append({
                "is_index": (child.get("IsIndexedRead") == "Y"),
                "nb_cycles": int(child.get("NumCycles"))
            })
        return reads

    def _getInstrumentFromRoot(self, root):
        serial_number = root.find("ScannerID")
        if serial_number is None:
            serial_number = root.find("InstrumentID")
        serial_number = serial_number.text
        return {
            "id": serial_number,
            "platform": platformFromInstrumentSerialNumber(serial_number)
        }

    def _getRunFromRoot(self, root):
        run_number = root.find("RunNumber")
        if run_number is None:
            run_number = root.find("ScanNumber")
        run_number = run_number.text.lstrip("0")
        return {
            "number": run_number,
            "id": root.find("RunID").text,
            "start_date": datetime.datetime.strptime(root.find("RunStartDate").text, '%y%m%d')
        }

    def _getPostProcessFromRoot(self, root):
        workflow = root.find("AnalysisWorkflowType")  # NextSeq style
        if workflow is None:
            workflow_markup = root.find("Workflow")  # MiSeq style
            if workflow_markup is not None:
                workflow = workflow_markup.find("Analysis")
            # Else HiSeq does not process any post-processing
        if workflow is not None:
            workflow = None if workflow.text == "" else workflow.text
        return workflow

    def _getKitFromRoot(self, root):
        # Flowcell ID
        flowcell_id = root.find("FlowCellSerial")  # NextSeq style
        if flowcell_id is not None:
            flowcell_id = flowcell_id.text
        else:
            flowcell_markup = root.find("FlowcellRFIDTag")  # MiSeq style
            if flowcell_markup is not None:
                flowcell_id = flowcell_markup.find("SerialNumber").text
            else:
                flowcell_id = root.find("Barcode")  # HiSeq style
                if flowcell_id is not None:
                    flowcell_id = flowcell_id.text
                else:
                    flowcell_id = self.run["id"].rsplit("_", 1)[1]
        # Reagent kit ID
        reagent_kit_id = root.find("ReagentKitBarcode")
        if reagent_kit_id is None:
            reagent_kit_id = root.find("ReagentKitSerial")
        if reagent_kit_id is not None:
            reagent_kit_id = reagent_kit_id.text
        return {
            "flowcell_id": flowcell_id,
            "reagent_kit_id": reagent_kit_id
        }

    def _getSoftwareFromRoot(self, root):
        cs_root_markup = root.find("Setup")
        if cs_root_markup is None:
            cs_root_markup = root
        return {
            "RTA": root.find("RTAVersion").text,
            "CS": cs_root_markup.find("ApplicationVersion").text
        }


class CompletedJobInfo(object):
    """Reader for CompletedJobInfo.xml."""

    def __init__(self, path, date_format='%Y-%m-%dT%H:%M:%S'):
        self.filepath = path
        self.date_format = date_format
        self.start_datetime = None
        self.end_datetime = None
        self.version = None
        self.workflow_name = None
        self.parameters = None
        self._parse()

    def _parse(self):
        # Retrieve information by section
        tree = ET.parse(self.filepath)
        root = tree.getroot()
        # Process information
        self.end_datetime = root.find("CompletionTime").text  # 2017-11-23T17:19:55.0941558+01:00
        self.end_datetime = self.end_datetime.split(".")[0]
        self.end_datetime = datetime.datetime.strptime(self.end_datetime, self.date_format)
        self.start_datetime = root.find("StartTime").text  # 2017-11-23T17:19:55.0941558+01:00
        self.start_datetime = self.start_datetime.split(".")[0]
        self.start_datetime = datetime.datetime.strptime(self.start_datetime, self.date_format)
        self.version = root.find("Workflow").find("WorkflowVersion").text
        self.workflow_name = root.find("Workflow").find("Analysis").text
        workflow_param = etreeToDict(root.find("Workflow").find("WorkflowSettings"))
        self.parameters = {"WorkflowSettings": workflow_param}
        if self.workflow_name == "Amplicon - DS":
            self.parameters["AmpliconSettings"] = etreeToDict(root.find("Workflow").find("AmpliconSettings"))


def etreeToDict(node):
    data = {"key": node.tag}
    # Attributes
    if len(node.attrib):
        data["attributes"] = node.attrib
    # Children
    children = list()
    for child in node:
        children.append(etreeToDict(child))
    if len(children):
        data["children"] = children
    else:
        data["val"] = node.text
    # Return
    return data


def getIlluminaName(name):
    """
    Return sample name used by Illumina in filename.

    :param name: The name provided to Illumina process (for example in samplesheet).
    :type name: str
    :return: The sample name used by Illumina as part of filename.
    :rtype: str
    """
    return name.replace("_", "-").replace(" ", "-").replace(".", "-").replace("+", "")


def getLibNameFromReadsPath(seq_path):
    """
    Return library name from the path of the sequences file.

    :param seq_path: The path of the sequences file.
    :type seq_path: str
    :return: The library name.
    :rtype: str
    """
    library_name, extensions = os.path.basename(seq_path).split(".", 1)
    for curr_ext in extensions.split("."):
        if curr_ext not in ["fq", "fastq", "fasta", "fa", "gz", "bz", "bz2", "lz", "zip"]:
            raise Exception('The file "{}" cannot be processed by getLibNameFromReadsPath because the extension "{}" is not managed.'.format(seq_path, curr_ext))
    if re.search('_[rR][1-2]$', library_name):
        library_name = library_name[:-3]
    elif re.search('_[rR][1-2]_\d\d\d$', library_name):
        library_name = library_name[:-7]
    return library_name


def getInfFromSeqID(seq_id):
    """
    Return sequencer id, run id, flowcell id and position of the sequence on the sequencer flowcell.

    :param sequence: The ID of the sequence provided by the sequencer.
    :type sequence: str
    :return: The sequencer id, run id, flowcell id and position of the sequence on the sequencer flowcell.
    :rtype: dict
    """
    fields = seq_id.split(":")
    if len(fields) == 7:  # Illumina's ID: EAS139:136:FC706VJ:2:2104:15343:197393
        fields.append(None)
    # else  Illumina's ID: EAS139:136:FC706VJ:2:2104:15343:197393:ATGCATA+CTAGC
    return {
        "sequencer_id": fields[0],
        "run_id": fields[1],
        "flowcell_id": fields[2],
        "lane_id": int(fields[3]),
        "tile_id": int(fields[4]),
        "x_pos": int(fields[5]),
        "y_pos": int(fields[6]),
        "umi": fields[7]
    }


def getInfFromSeqDesc(seq_desc):
    """
    Return sequencer id, run id, flowcell id and position of the sequence on the sequencer flowcell.

    :param sequence: The ID of the sequence provided by the sequencer.
    :type sequence: str
    :return: The sequencer id, run id, flowcell id and position of the sequence on the sequencer flowcell.
    :rtype: dict
    """
    # Illumina's description: 1:Y:18:ATCACG
    reads_phases, kept_status, control_bits, barcode = seq_desc.split(":")
    return {
        "reads_phases": int(reads_phases),
        "is_kept": kept_status == "N",
        "control_bits": None if control_bits == "0" else int(control_bits),
        "barcode": None if barcode == "" else barcode
    }


def platformFromInstrumentSerialNumber(instrument_id):
    """
    Return platform name from instrument ID.

    :param instrument_id: The instrument serial number.
    :type instrument_id: str
    :return: The platform name (Hiseq or NextSeq or ...).
    :rtype: str
    """
    platform_by_re = {
        # "?": "iSeq",
        "^MN[0-9]{5}$": "MiniSeq",
        "^ML-..-[0-9]{2}$": "MiniSeq",
        "^M[0-9]{5}$": "MiSeq",
        "^N[SB][0-9]{6}$": "NextSeq",
        "^NDX[0-9]{6}": "NextSeq",
        "^[CDJKE][0-9]{5}$": "HiSeq",
        "^A[0-9]{5}$": "NovaSeq"
    }
    if instrument_id.startswith("HWI"):
        instrument_id = instrument_id[3:]
    platform = None
    for curr_re, curr_instru in platform_by_re.items():
        if platform is None:
            if re.search(curr_re, instrument_id):
                platform = curr_instru
    return platform


def getRunFolderInfo(run_folder):
    """
    Return run information (instrument, run configuration, samples, ...) coming from several file in run folder.

    :param run_folder: Path to the run folder.
    :type run_folder: str
    :return: Run information.
    :rtype: dict
    """
    # RunInfo
    run_info_path = os.path.join(run_folder, "RunInfo.xml")
    run_info = RunInfo(run_info_path)
    run = {
        "reads_phases": run_info.reads_phases,
        "flowcell": run_info.flowcell,
        "instrument": run_info.instrument,
        "info": run_info.run,
        "samples": None
    }
    run["info"]["end_date"] = None
    # RTAComplete
    rta_complete_path = os.path.join(run_folder, "RTAComplete.txt")
    if os.path.exists(rta_complete_path):
        rta_complete = RTAComplete(rta_complete_path)
        run["info"]["end_date"] = rta_complete.end_date
    # SampleSheet
    samplesheet_path = os.path.join(run_folder, "SampleSheet.csv")
    if os.path.exists(samplesheet_path):
        run["samples"] = SampleSheetIO(samplesheet_path).samples
    return run
